import itertools
import numpy as np
import pandas as pd
import sys

from westpa.core.binning.assign import BinMapper
from westpa.core.h5io import WESTPAH5File, tostr
from westpa.core.segment import Segment
from westpa.core.states import BasisState, InitialState, TargetState
from westpa.tools.binning import mapper_from_hdf5


class Run:
    """A read-only view of a WESTPA simulation run.

    Parameters
    ----------
    h5filename : str or file-like object, default 'west.h5'
        Pathname or stream of a main WESTPA HDF5 data file.

    """

    DESCRIPTION = 'WESTPA Run'

    def __init__(self, h5filename='west.h5'):
        self.h5filename = h5filename

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, traceback):
        self.close()

    @classmethod
    def open(cls, h5filename='west.h5'):
        """Alternate constructor.

        Parameters
        ----------
        h5filename : str or file-like object, default 'west.h5'
            Pathname or stream of a main WESTPA HDF5 data file.

        """
        return cls(h5filename)

    def close(self):
        """Close the Run instance by closing the underlying WESTPA HDF5 file."""
        self.h5file.close()

    @property
    def closed(self):
        """bool: Whether the Run instance is closed."""
        return not bool(self.h5file)

    @property
    def h5filename(self):
        return self.h5file.filename

    @h5filename.setter
    def h5filename(self, value):
        try:
            h5file = WESTPAH5File(value, 'r')
        except FileNotFoundError as e:
            e.strerror = f'Failed to open {self.DESCRIPTION}: file {value!r} not found'
            raise e.with_traceback(None)
        self.h5file = h5file

    @property
    def summary(self):
        """pd.DataFrame: Summary data by iteration."""
        df = pd.DataFrame(
            self.h5file['summary'][: self.num_iterations],
            index=range(1, self.num_iterations + 1),
            dtype=object,
        )
        df.pop('norm')  # should always be 1.0
        df.pop('binhash')  # not human readable
        return df

    @property
    def num_iterations(self):
        """int: Number of completed iterations."""
        if not hasattr(self, '_num_iterations'):
            current = self.h5file.attrs['west_current_iteration']
            grp = self.h5file.get_iter_group(current)
            if (grp['seg_index']['status'] == Segment.SEG_STATUS_COMPLETE).all():
                self._num_iterations = current
            else:
                self._num_iterations = current - 1
        return self._num_iterations

    @property
    def iterations(self):
        """Sequence[Iteration]: Sequence of iterations."""
        return [Iteration(number, self) for number in range(1, self.num_iterations + 1)]

    @property
    def num_walkers(self):
        """int: Total number of walkers."""
        return sum(iteration.num_segments for iteration in self)

    @property
    def num_segments(self):
        """int: Total number of trajectory segments (alias self.num_walkers)."""
        return self.num_walkers

    @property
    def walkers(self):
        """Iterable[Walker]: All walkers in the run."""
        return itertools.chain.from_iterable(self)

    @property
    def recycled_walkers(self):
        """Iterable[Walker]: Walkers that stopped in the sink."""
        return itertools.chain.from_iterable(iteration.recycled_walkers for iteration in self)

    @property
    def initial_walkers(self):
        """Iterable[Walker]: Walkers whose parents are initial states."""
        return itertools.chain.from_iterable(iteration.initial_walkers for iteration in self)

    def iteration(self, number):
        """Return a specific iteration.

        Parameters
        ----------
        number : int
            Iteration number (1-based).

        Returns
        -------
        Iteration
            The iteration indexed by `number`.

        """
        valid_range = range(1, self.num_iterations + 1)
        if number not in valid_range:
            raise ValueError(f'iteration number must be in {valid_range}')
        return Iteration(number, self)

    def __len__(self):
        return self.num_iterations

    def __iter__(self):
        return iter(self.iterations)

    def __contains__(self, iteration):
        return iteration.run == self

    def __eq__(self, other):
        return self.h5file == other.h5file

    def __hash__(self):
        return hash(self.h5file)

    def __bool__(self):
        return not self.closed

    def __repr__(self):
        if self.closed:
            return f'<Closed {self.DESCRIPTION} at {hex(id(self))}>'
        return f'<{self.DESCRIPTION} with {self.num_iterations} iterations at {hex(id(self))}>'


class Iteration:
    """An iteration of a WESTPA simulation.

    Parameters
    ----------
    number : int
        Iteration number (1-based).
    run : Run
        Simulation run to which the iteration belongs.

    """

    def __init__(self, number, run):
        self.number = number
        self.run = run

    @property
    def h5group(self):
        """h5py.Group: HDF5 group containing the iteration data."""
        return self.run.h5file.get_iter_group(self.number)

    @property
    def prev(self):
        """Iteration: Previous iteration."""
        if self.number == 1:
            return None
        return self.run.iteration(self.number - 1)

    @property
    def next(self):
        """Iteration: Next iteration."""
        if self.number == self.run.num_iterations:
            return None
        return self.run.iteration(self.number + 1)

    @property
    def summary(self):
        """pd.DataFrame: Iteration summary."""
        df = pd.DataFrame(
            self.run.h5file['summary'][[self.number - 1]],
            index=[self.number],
            dtype=object,
        )
        df.pop('norm')  # should always be 1.0
        df.pop('binhash')  # not human readable
        return df.iloc[0]

    @property
    def segment_summaries(self):
        """pd.DataFrame: Segment summary data for the iteration."""
        df = pd.DataFrame(self.h5group['seg_index'][:], dtype=object)

        # Make 'endpoint_type' and 'status' human-readable.
        names = map(Segment.endpoint_type_names.get, df['endpoint_type'])
        df['endpoint_type'] = [name.split('_')[-1] for name in names]
        names = map(Segment.status_names.get, df['status'])
        df['status'] = [name.split('_')[-1] for name in names]

        return df

    @property
    def pcoords(self):
        """3D ndarray: Progress coordinate snaphots of each walker."""
        return self.h5group['pcoord'][:]

    @property
    def weights(self):
        """1D ndarray: Statistical weight of each walker."""
        return self.h5group['seg_index']['weight']

    @property
    def bin_target_counts(self):
        """1D ndarray, dtype=uint64: Target count for each bin."""
        val = self.h5group.get('bin_target_counts')
        if val is None:
            return None
        return val[:]

    @property
    def bin_mapper(self):
        """BinMapper: Bin mapper used in the iteration."""
        if self.bin_target_counts is None:
            return None
        mapper, _, _ = mapper_from_hdf5(self.run.h5file['bin_topologies'], self.h5group.attrs['binhash'])
        return mapper

    @property
    def num_bins(self):
        """int: Number of bins."""
        if self.number == 1:
            return 1
        return self.bin_target_counts.shape[0]

    @property
    def bins(self):
        """Iterable[Bin]: Bins."""
        mapper = self.bin_mapper
        return (Bin(index, mapper) for index in range(self.num_bins))

    @property
    def num_walkers(self):
        """int: Number of walkers in the iteration."""
        return self.h5group['seg_index'].shape[0]

    @property
    def num_segments(self):
        """int: Number of trajectory segments (alias self.num_walkers)."""
        return self.num_walkers

    @property
    def walkers(self):
        """Iterable[Walker]: Walkers in the iteration."""
        return (Walker(index, self) for index in range(self.num_walkers))

    @property
    def recycled_walkers(self):
        """Iterable[Walker]: Walkers that stopped in the sink."""
        endpoint_type = self.h5group['seg_index']['endpoint_type']
        indices = np.flatnonzero(endpoint_type == Segment.SEG_ENDPOINT_RECYCLED)
        return (Walker(index, self) for index in indices)

    @property
    def initial_walkers(self):
        """Iterable[Walker]: Walkers whose parents are initial states."""
        parent_ids = self.h5group['seg_index']['parent_id']
        return (walker for walker, parent_id in zip(self, parent_ids) if parent_id < 0)

    @property
    def auxiliary_data(self):
        """h5py.Group or None: Auxiliary data stored for the iteration."""
        return self.h5group.get('auxdata')

    @property
    def basis_state_summaries(self):
        """pd.DataFrame: Basis state summary data."""
        df = pd.DataFrame(self.h5group['ibstates']['bstate_index'][:])
        df['label'] = tostr(df['label'].str)
        df['auxref'] = tostr(df['auxref'].str)
        return df

    @property
    def basis_state_pcoords(self):
        """2D ndarray: Progress coordinates of each basis state."""
        return self.h5group['ibstates']['bstate_pcoord'][:]

    @property
    def basis_states(self):
        """list[BasisState]: Basis states in use for the iteration."""
        return [
            BasisState(row.label, row.probability, pcoord=pcoord, auxref=row.auxref, state_id=row.Index)
            for row, pcoord in zip(self.basis_state_summaries.itertuples(), self.basis_state_pcoords)
        ]

    @property
    def has_target_states(self):
        """bool: Whether target (sink) states are defined for this iteration."""
        return 'tstates' in self.h5group

    @property
    def target_state_summaries(self):
        """pd.DataFrame or None: Target state summary data."""
        if self.has_target_states:
            df = pd.DataFrame(self.h5group['tstates']['index'][:])
            df['label'] = tostr(df['label'].str)

            return df
        else:
            return None

    @property
    def target_state_pcoords(self):
        """2D ndarray or None: Progress coordinates of each target state."""
        return self.h5group['tstates']['pcoord'][:] if self.has_target_states else None

    @property
    def target_states(self):
        """list[TargetState]: Target states in use for the iteration."""
        if not self.has_target_states:
            return []
        return [
            TargetState(row.label, pcoord, state_id=row.Index)
            for row, pcoord in zip(self.target_state_summaries.itertuples(), self.target_state_pcoords)
        ]

    @property
    def sink(self):
        """BinUnion or None: Union of bins serving as the recycling sink."""
        if not self.has_target_states:
            return None
        mapper = Iteration(self.number + 1, self.run).bin_mapper
        return BinUnion(mapper.assign(self.target_state_pcoords), mapper)

    def bin(self, index):
        """Return the bin with the given index.

        Parameters
        ----------
        index : int
            Bin index (0-based).

        Returns
        -------
        Bin
            The bin indexed by `index`.

        """
        return Bin(index, self.bin_mapper)

    def walker(self, index):
        """Return the walker with the given index.

        Parameters
        ----------
        index : int
            Walker index (0-based).

        Returns
        -------
        Walker
            The walker indexed by `index`.

        """
        valid_range = range(self.num_walkers)
        if index not in valid_range:
            raise ValueError(f'walker index must be in {valid_range}')
        return Walker(index, self)

    def basis_state(self, index):
        """Return the basis state with the given index.

        Parameters
        ----------
        index : int
            Basis state index (0-based).

        Returns
        -------
        BasisState
            The basis state indexed by `index`.

        """
        row = self.h5group['ibstates']['bstate_index'][index]
        pcoord = self.h5group['ibstates']['bstate_pcoord'][index]
        return BasisState(
            tostr(row['label']),
            row['probability'],
            pcoord=pcoord,
            auxref=tostr(row['auxref']),
            state_id=index,
        )

    def target_state(self, index):
        """Return the target state with the given index.

        Parameters
        ----------
        index : int
            Target state index (0-based).

        Returns
        -------
        TargetState
            The target state indexed by `index`.

        """
        row = self.h5group['tstates']['index'][index]
        pcoord = self.h5group['tstates']['pcoord'][index]
        return TargetState(tostr(row['label']), pcoord, state_id=index)

    def __iter__(self):
        return iter(self.walkers)

    def __contains__(self, walker):
        return walker.iteration == self

    def __eq__(self, other):
        return self.number == other.number and self.run == other.run

    def __hash__(self):
        return hash((self.number, self.run))

    def __repr__(self):
        return f'{self.__class__.__name__}({self.number}, {self.run})'


class Walker:
    """A walker in an iteration of a WESTPA simulation.

    Parameters
    ----------
    index : int
        Walker index (0-based).
    iteration : Iteration
        Iteration to which the walker belongs.

    """

    def __init__(self, index, iteration):
        self.index = index
        self.iteration = iteration

    @property
    def run(self):
        """Run: Run to which the walker belongs."""
        return self.iteration.run

    @property
    def weight(self):
        """float64: Statistical weight of the walker."""
        return self.iteration.weights[self.index]

    @property
    def pcoords(self):
        """2D ndarray: Progress coordinate snapshots."""
        return self.iteration.h5group['pcoord'][self.index]

    @property
    def num_snapshots(self):
        """int: Number of snapshots."""
        return self.pcoords.shape[0]

    @property
    def segment_summary(self):
        """pd.Series: Segment summary data."""
        df = pd.DataFrame(
            self.iteration.h5group['seg_index'][[self.index]],
            index=[self.index],
            dtype=object,
        )

        # Make 'endpoint_type' and 'status' human-readable.
        names = map(Segment.endpoint_type_names.get, df['endpoint_type'])
        df['endpoint_type'] = [name.split('_')[-1] for name in names]
        names = map(Segment.status_names.get, df['status'])
        df['status'] = [name.split('_')[-1] for name in names]

        return df.iloc[0]

    @property
    def parent(self):
        """Walker or InitialState: The parent of the walker."""
        parent_id = self.iteration.h5group['seg_index']['parent_id'][self.index]

        if parent_id >= 0:
            return Walker(parent_id, self.iteration.prev)

        istate_id = -(parent_id + 1)
        row = self.iteration.h5group['ibstates']['istate_index'][istate_id]

        # Initial states may or may not be generated from a basis state.
        bstate_id = row['basis_state_id']
        try:
            bstate = self.iteration.basis_state(bstate_id)
        except IndexError:
            bstate = None
            bstate_id = None

        return InitialState(
            istate_id,
            bstate_id,
            row['iter_created'],
            iter_used=row['iter_used'],
            istate_type=row['istate_type'],
            istate_status=row['istate_status'],
            pcoord=self.iteration.h5group['ibstates']['istate_pcoord'][istate_id],
            basis_state=bstate,
        )

    @property
    def children(self):
        """Iterable[Walker]: The children of the walker."""
        next = self.iteration.next
        if next is None:
            return ()
        indices = np.flatnonzero(next.h5group['seg_index']['parent_id'] == self.index)
        return (Walker(index, next) for index in indices)

    @property
    def recycled(self):
        """bool: True if the walker stopped in the sink, False otherwise."""
        endpoint_type = self.iteration.h5group['seg_index']['endpoint_type'][self.index]
        return endpoint_type == Segment.SEG_ENDPOINT_RECYCLED

    @property
    def initial(self):
        """bool: True if the parent of the walker is an initial state, False otherwise."""
        return self.iteration.h5group['seg_index']['parent_id'][self.index] < 0

    @property
    def auxiliary_data(self):
        """dict: Auxiliary data for the walker."""
        data = self.iteration.auxiliary_data or {}
        return {name: data[name][self.index] for name in data}

    def trace(self, **kwargs):
        """Return the trace (ancestral line) of the walker.

        For full documentation see :class:`Trace`.

        Returns
        -------
        Trace
            The trace of the walker.

        """
        return Trace(self, **kwargs)

    def __eq__(self, other):
        return self.index == other.index and self.iteration == other.iteration

    def __hash__(self):
        return hash((self.index, self.iteration))

    def __repr__(self):
        return f'{self.__class__.__name__}({self.index}, {self.iteration})'


class BinUnion:
    """A (disjoint) union of bins defined by a common bin mapper.

    Parameters
    ----------
    indices : iterable of int
        The indices of the bins comprising the union.
    mapper : BinMapper
        The bin mapper defining the bins.

    """

    def __init__(self, indices, mapper):
        if not isinstance(mapper, BinMapper):
            raise TypeError(f'mapper must be an instance of {BinMapper}')

        indices = set(indices)
        valid_range = range(mapper.nbins)
        if any(index not in valid_range for index in indices):
            raise ValueError(f'bin indices must be in {valid_range}')

        self.indices = indices
        self.mapper = mapper

    def union(self, *others):
        """Return the union of the bin union and all others.

        Parameters
        ----------
        *others : BinUnion
            Other :class:`BinUnion` instances, consisting of bins defined by
            the same underlying bin mapper.

        Returns
        -------
        BinUnion
            The union of `self` and `others`.

        """
        if any(other.mapper != self.mapper for other in others):
            raise ValueError('bins must be defined by the same bin mapper')
        indices = self.indices.union(*(other.indices for other in others))
        return BinUnion(indices, self.mapper)

    def intersection(self, *others):
        """Return the intersection of the bin union and all others.

        Parameters
        ----------
        *others : BinUnion
            Other :class:`BinUnion` instances, consisting of bins defined by
            the same underlying bin mapper.

        Returns
        -------
        BinUnion
            The itersection of `self` and `others`.

        """
        if any(other.mapper != self.mapper for other in others):
            raise ValueError('bins must be defined by the same bin mapper')
        indices = self.indices.intersection(*(other.indices for other in others))
        return BinUnion(indices, self.mapper)

    def __contains__(self, coord):
        result = self.mapper.assign([coord])
        if result.size != 1:
            raise ValueError('left operand must be a single coordinate tuple')
        return result[0] in self.indices

    def __or__(self, other):
        return self.union(other)

    def __and__(self, other):
        return self.intersection(other)

    def __bool__(self):
        return bool(self.indices)

    def __repr__(self):
        return f'{self.__class__.__name__}({self.indices}, {self.mapper})'


class Bin(BinUnion):
    """A bin defined by a bin mapper.

    Parameters
    ----------
    index : int
        The index of the bin.
    mapper : BinMapper
        The bin mapper defining the bin.

    """

    def __init__(self, index, mapper):
        super().__init__({index}, mapper)
        self.index = index

    def __repr__(self):
        return f'{self.__class__.__name__}({self.index}, {self.mapper})'


class Trace:
    """A trace of a walker's ancestry.

    Parameters
    ----------
    walker : Walker
        The terminal walker.
    source : Bin, BinUnion, or collections.abc.Container, optional
        A source (macro)state, specified as a container object whose
        :meth:`__contains__` method is the indicator function for the
        corresponding subset of progress coordinate space. The trace is
        stopped upon encountering a walker that stopped in `source`.
    max_length : int, optional
        The maximum number of walkers in the trace.

    """

    def __init__(self, walker, source=None, max_length=None):
        if max_length is None:
            max_length = sys.maxsize
        else:
            max_length = int(max_length)
            if max_length < 1:
                raise ValueError('max_length must be at least 1')

        walkers = []
        initial_state = None
        while len(walkers) < max_length:
            if source and walker.pcoords[-1] in source:
                break
            walkers.append(walker)
            parent = walker.parent
            if isinstance(parent, InitialState):
                initial_state = parent
                break
            walker = parent
        walkers.reverse()

        self.walkers = walkers
        self.initial_state = initial_state
        self.source = source
        self.max_length = max_length

    def __len__(self):
        return len(self.walkers)

    def __iter__(self):
        return iter(self.walkers)

    def __contains__(self, walker):
        return walker in self.walkers

    def __getitem__(self, key):
        return self.walkers[key]

    def __repr__(self):
        s = f'Trace({self.walkers[-1]}'
        if self.source:
            s += f', source={self.source}'
        if self.max_length < sys.maxsize:
            s += f', max_length={self.max_length}'
        return s + ')'



import logging
import os
import re

import h5py
import numpy as np

from westpa.tools import WESTMasterCommand, WESTSubcommand, ProgressIndicatorComponent, Plotter
from westpa.core import h5io

if os.environ.get('DISPLAY') is not None:
    from matplotlib import pyplot

log = logging.getLogger('ploterrs')


class CommonPloterrs(WESTSubcommand):
    def __init__(self, parent):
        super().__init__(parent)

        self.progress = ProgressIndicatorComponent()

        self.xscale = None
        self.yscale = None
        self.xrange = None
        self.yrange = None
        self.xlabel = None
        self.ylabel = None
        self.title = None

        self.plot_options_group = None

    def add_args(self, parser):
        self.progress.add_args(parser)

        pogroup = self.plot_options_group = parser.add_argument_group('plot options')
        pogroup.add_argument(
            '--xscale',
            choices=['linear', 'log', 'symlog'],
            default='linear',
            help='''Use "linear", "log", or "symlog" scaling for the x axis.
                             (Default: %(default)s).''',
        )
        pogroup.add_argument(
            '--yscale',
            choices=['linear', 'log', 'symlog'],
            default='linear',
            help='''Use "linear", "log", or "symlog" scaling for the y axis.
                             (Default: %(default)s).''',
        )
        pogroup.add_argument(
            '--xrange',
            help='''Restrict X range to XRANGE, which must be formatted as "xmin,xmax".
                             (Default: determined by input data.)''',
        )
        pogroup.add_argument(
            '--yrange',
            help='''Restrict Y range to YRANGE, which must be formatted as "ymin,ymax".
                             (Default: determined by input data.)''',
        )
        pogroup.add_argument('--xlabel', help='''Use XLABEL for the x-axis label. (Default: varies.)''')
        pogroup.add_argument('--ylabel', help='''Use YLABEL for the y-axis label. (Default: varies.)''')
        pogroup.add_argument('--title', help='''Use TITLE for the plot title. (Default: varies.)''')
        pogroup.add_argument('--terminal', '-t', dest='plotting', action='store_true', help='''Plot output in terminal.''')

    def process_args(self, args):
        self.progress.process_args(args)

        if args.xrange:
            self.xrange = self.parse_range(args.xrange)

        if args.yrange:
            self.yrange = self.parse_range(args.yrange)

        self.xscale = args.xscale
        self.yscale = args.yscale
        self.xlabel = args.xlabel or 'Iteration'
        self.ylabel = args.ylabel
        self.title = args.title
        if args.plotting or os.environ.get('DISPLAY') is None:
            self.interface = 'text'
        else:
            self.interface = 'matplotlib'

    def parse_range(self, rangespec):
        try:
            (lbt, ubt) = rangespec.split(',')
            return float(lbt), float(ubt)
        except (ValueError, TypeError) as e:
            raise ValueError('invalid range specification {!r}: {!s}'.format(rangespec, e))

    def do_plot(self, data, output_filename, title=None, x_range=None, y_range=None, x_label=None, y_label=None):
        if not output_filename:
            return

        title = title or self.title
        x_range = x_range or self.xrange
        y_range = y_range or self.yrange
        x_label = x_label or self.xlabel
        y_label = y_label or self.ylabel

        iters = data['iter_stop'] - 1

        pyplot.figure()
        pyplot.plot(iters, data['expected'], color='black')
        pyplot.plot(iters, data['ci_lbound'], color='gray')
        pyplot.plot(iters, data['ci_ubound'], color='gray')

        pyplot.gca().set_xscale(self.xscale)
        pyplot.gca().set_yscale(self.yscale)

        if title:
            pyplot.title(title)

        if x_range is not None:
            pyplot.xlim(x_range)

        if y_range is not None:
            pyplot.ylim(y_range)

        if x_label:
            pyplot.xlabel(x_label)

        if y_label:
            pyplot.ylabel(y_label)

        pyplot.savefig(output_filename)


class GenericIntervalSubcommand(CommonPloterrs):
    description = '''\
Plots generic expectation/CI data. A path to the HDF5 file and the dataset
within it must be provided. This path takes the form **FILENAME/PATH[SLICE]**.
If the dataset is not a vector (one dimensional) then a slice must be provided.
For example, to access the state 0 to state 1 rate evolution calculated by
``w_kinavg``, one would use ``kinavg.h5/rate_evolution[:,0,1]``.


-----------------------------------------------------------------------------
Command-line arguments
-----------------------------------------------------------------------------
'''
    subcommand = 'generic'
    help_text = 'arbitrary HDF5 file and dataset'

    def __init__(self, parent):
        super().__init__(parent)
        self.h5file = None
        self.h5dset = None
        self.dset_slice = None
        self.output_filename = None

    def add_args(self, parser):
        iogroup = parser.add_argument_group('input/output options')
        iogroup.add_argument(
            '-o',
            '--output',
            default='errbars.pdf',
            help='''Write plot to OUTPUT (default: %(default)s), whose format will
                             be determined by filename extension.''',
        )
        iogroup.add_argument(
            'dsspec',
            help='''Use data located at DSSPEC, which must be formatted as
                             FILENAME/PATH[SLICE]. FILENAME is the HDF5 file to read, PATH is the
                             HDF5 path to the dataset, and SLICE, if provided, must be the Numpy-style
                             slice (including brackets) which selects a vector of data of the
                             appropriate type.''',
        )

    def process_args(self, args):
        self.output_filename = args.output
        (pathname, slicestr) = re.search(r'([^[]+)(\[[^\]]+\])?$', args.dsspec).groups()
        if slicestr:
            sl = eval('np.index_exp' + slicestr)
        else:
            sl = np.index_exp[...]
        self.h5file, self.h5dset = h5io.resolve_filepath(pathname, mode='r')
        self.dset_slice = sl

    def load_and_validate_data(self):
        reqd_fields = set(['iter_start', 'iter_stop', 'expected', 'ci_lbound', 'ci_ubound'])

        self.progress.indicator.new_operation('loading data')
        data = self.h5dset[self.dset_slice]

        if data.ndim != 1:
            raise TypeError('dataset to be plotted must be 1-dimensional')
        try:
            fieldnames = set(data.dtype.fields.keys())
        except AttributeError:
            raise TypeError('dataset has inappropriate type')
        else:
            if len(fieldnames & reqd_fields) < len(reqd_fields):
                raise TypeError('dataset does not contain correct fields')

        return data

    def go(self):
        with self.progress.indicator:
            data = self.load_and_validate_data()
            self.progress.indicator.new_operation('plotting')
            self.do_plot(data, self.output_filename)


class DirectKinetics(CommonPloterrs):
    subcommand = 'd.kinetics'
    help_text = 'output of w_direct kinetics'
    input_filename = 'direct.h5'
    flux_output_filename = 'flux_evolution_d_{state_label}.pdf'
    rate_output_filename = 'rate_evolution_d_{istate_label}_{fstate_label}.pdf'
    description = '''\
Plot evolution of state-to-state rates and total flux into states as generated
by ``w_{direct/reweight} kinetics`` (when used with the ``--evolution-mode``
option). Plots are generated for all rates/fluxes calculated. Output filenames
require (and plot titles and axis labels support) substitution based on which
flux/rate is being plotted:

  istate_label, fstate_label
    *(String, for rates)* Names of the initial and final states, as originally
    given to ``w_assign``.

  istate_index, fstate_index
    *(Integer, for rates)* Indices of initial and final states.

  state_label
    *(String, for fluxes)* Name of state

  state_index
    *(Integer, for fluxes)* Index of state
'''

    def __init__(self, parent):
        super().__init__(parent)
        self.kinavg_file = None

        self.dset_slice = None
        self.rate_output_pattern = None
        self.flux_output_pattern = None

        self.state_labels = None

    def add_args(self, parser):
        iogroup = parser.add_argument_group('input/output')
        iogroup.add_argument(
            '-i', '--input', default=self.input_filename, help='''Read kinetics results from INPUT (default: %(default)s).'''
        )
        iogroup.add_argument(
            '--rate-output',
            default=self.rate_output_filename,
            help='''Filename pattern for rate evolution output. See above for valid
                             field names. (Default: %(default)r).''',
        )
        iogroup.add_argument(
            '--flux-output',
            default=self.flux_output_filename,
            help='''Filename pattern for flux evolution output. See above for valid
                             field names. (Default: %(default)r).''',
        )

    def process_args(self, args):
        self.kinavg_file = h5py.File(args.input, 'r')

        self.state_labels = list(self.kinavg_file['state_labels'][...])

        self.rate_output_pattern = args.rate_output
        self.flux_output_pattern = args.flux_output

    def plot_flux(self, istate):
        label = self.state_labels[istate]
        data = self.kinavg_file['target_flux_evolution'][:, istate]

        if (data['iter_start'] == 0).all():
            # No data
            return

        subdict = dict(state_label=label, state_index=istate)

        output_filename = self.flux_output_pattern.format(**subdict) if self.flux_output_pattern else None

        title = self.title if self.title is not None else 'Flux into state "{state_label}"'
        title = title.format(**subdict)

        x_label = self.xlabel.format(**subdict) if self.xlabel else None

        y_label = self.ylabel if self.ylabel is not None else r'Flux $(\tau^{{-1}})$'
        y_label = y_label.format(**subdict)

        self.do_plot(data, output_filename, title, x_label=x_label, y_label=y_label)

    def plot_rate(self, istate, jstate):
        ilabel = self.state_labels[istate]
        jlabel = self.state_labels[jstate]
        data = self.kinavg_file['rate_evolution'][:, istate, jstate]

        if (data['iter_start'] == 0).all():
            # No data
            return

        subdict = dict(istate_label=ilabel, istate_index=istate, fstate_label=jlabel, fstate_index=jstate)

        output_filename = self.rate_output_pattern.format(**subdict) if self.rate_output_pattern else None

        title = self.title if self.title is not None else 'Rate from state "{istate_label}" to state "{fstate_label}"'
        title = title.format(**subdict)

        x_label = self.xlabel.format(**subdict) if self.xlabel else None

        y_label = self.ylabel if self.ylabel is not None else r'Rate $(\tau^{{-1}})$'
        y_label = y_label.format(**subdict)

        self.do_plot(data, output_filename, title, x_label=x_label, y_label=y_label)

    def go(self):
        pi = self.progress.indicator
        nstates = len(self.state_labels)
        if self.interface == 'matplotlib':
            with pi:
                # if --evolution-mode wasn't specified, neither of these exist:
                if 'target_flux_evolution' in self.kinavg_file:
                    pi.new_operation('plotting fluxes', nstates)
                    for istate in range(nstates):
                        self.plot_flux(istate)
                        pi.progress += 1

                # if --evolution-mode wasn't specified, we won't get this either
                if 'rate_evolution' in self.kinavg_file:
                    pi.new_operation('plotting rates', nstates * nstates)
                    for istate in range(nstates):
                        for jstate in range(nstates):
                            self.plot_rate(istate, jstate)
                            pi.progress += 1
                else:
                    print('rate evolution not available')
        else:
            plotter = Plotter(self.kinavg_file, 'rate_evolution', iteration=-1, interface='text')
            for istate in range(nstates):
                for jstate in range(nstates):
                    if istate != jstate:
                        plotter.plot(istate, jstate)
            plotter = Plotter(self.kinavg_file, 'conditional_flux_evolution', iteration=-1, interface='text')
            for istate in range(nstates):
                for jstate in range(nstates):
                    if istate != jstate:
                        plotter.plot(istate, jstate)


class DirectStateprobs(CommonPloterrs):
    subcommand = 'd.probs'
    help_text = 'output of w_direct probs'
    input_filename = 'direct.h5'
    pop_output_filename = 'pop_evolution_d_{state_label}.pdf'
    color_output_filename = 'color_evolution_d_{state_label}.pdf'
    description = '''\
Plot evolution of macrostate populations and associated uncertainties. Plots
are generated for all states calculated. Output filenames require (and plot
titles and axis labels support) substitution based on which state is being
plotted:

  state_label
    *(String, for fluxes)* Name of state

  state_index
    *(Integer, for fluxes)* Index of state
'''

    def __init__(self, parent):
        super().__init__(parent)
        self.stateprobs_file = None

        self.dset_slice = None
        self.rate_output_pattern = None
        self.flux_output_pattern = None

        self.state_labels = None

    def add_args(self, parser):
        iogroup = parser.add_argument_group('input/output')
        iogroup.add_argument(
            '-i', '--input', default=self.input_filename, help='''Read w_kinavg results from INPUT (default: %(default)s).'''
        )
        iogroup.add_argument(
            '--population-output',
            default=self.pop_output_filename,
            help='''Filename pattern for population evolution output. See above for valid
                             field names. (Default: %(default)r).''',
        )
        iogroup.add_argument(
            '--color-output',
            default=self.color_output_filename,
            help='''Filename pattern for ensemble evolution output. See above for valid
                             field names. (Default: %(default)r).''',
        )

    def process_args(self, args):
        self.stateprobs_file = h5py.File(args.input, 'r')

        self.state_labels = list(self.stateprobs_file['state_labels'][...])
        self.pop_output_pattern = args.population_output
        self.color_output_pattern = args.color_output

    def plot_pop(self, istate):
        label = self.state_labels[istate]
        data = self.stateprobs_file['state_pop_evolution'][:, istate]

        if (data['iter_start'] == 0).all():
            # No data
            return

        subdict = dict(state_label=label, state_index=istate)

        output_filename = self.pop_output_pattern.format(**subdict) if self.pop_output_pattern else None

        title = self.title if self.title is not None else 'Population in state "{state_label}"'
        title = title.format(**subdict)

        x_label = self.xlabel.format(**subdict) if self.xlabel else None
        y_label = self.ylabel if self.ylabel is not None else r'Population'
        y_label = y_label.format(**subdict)

        self.do_plot(data, output_filename, title, x_label=x_label, y_label=y_label)

    def plot_color(self, istate):
        label = self.state_labels[istate]
        data = self.stateprobs_file['color_prob_evolution'][:, istate]

        if (data['iter_start'] == 0).all():
            # No data
            return

        subdict = dict(state_label=label, state_index=istate)

        output_filename = self.color_output_pattern.format(**subdict) if self.color_output_pattern else None

        title = self.title if self.title is not None else 'Population in ensemble "{state_label}"'
        title = title.format(**subdict)

        x_label = self.xlabel.format(**subdict) if self.xlabel else None
        y_label = self.ylabel if self.ylabel is not None else r'Population'
        y_label = y_label.format(**subdict)

        self.do_plot(data, output_filename, title, x_label=x_label, y_label=y_label)

    def go(self):
        pi = self.progress.indicator
        nstates = len(self.state_labels)
        if self.interface == 'matplotlib':
            with pi:
                if 'state_pop_evolution' in self.stateprobs_file:
                    pi.new_operation('plotting populations', nstates)
                    for istate in range(nstates):
                        self.plot_pop(istate)
                        pi.progress += 1

                if 'color_prob_evolution' in self.stateprobs_file:
                    pi.new_operation('plotting ensemble populations', nstates)
                    for istate in range(nstates):
                        self.plot_color(istate)
                        pi.progress += 1
                else:
                    print('population evolution not available')
        else:
            plotter = Plotter(self.stateprobs_file, 'state_pop_evolution', iteration=-1, interface='text')
            for istate in range(nstates):
                plotter.plot(istate)
            plotter = Plotter(self.stateprobs_file, 'color_prob_evolution', iteration=-1, interface='text')
            for istate in range(nstates):
                plotter.plot(istate)


class ReweightStateprobs(DirectStateprobs):
    subcommand = 'rw.probs'
    help_text = 'output of w_reweight probs'
    input_filename = 'reweight.h5'
    pop_output_filename = 'pop_evolution_rw_{state_label}.pdf'
    color_output_filename = 'color_evolution_rw_{state_label}.pdf'


class ReweightKinetics(DirectKinetics):
    subcommand = 'rw.kinetics'
    help_text = 'output of w_reweight kinetics'
    input_filename = 'reweight.h5'
    flux_output_filename = 'flux_evolution_rw_{state_label}.pdf'
    rate_output_filename = 'rate_evolution_rw_{istate_label}_{fstate_label}.pdf'


class PloterrsTool(WESTMasterCommand):
    prog = 'ploterrs'
    subcommands = [DirectKinetics, DirectStateprobs, ReweightStateprobs, ReweightKinetics, GenericIntervalSubcommand]
    subparsers_title = 'supported input formats'
    description = '''\
Plots error ranges for weighted ensemble datasets.


-----------------------------------------------------------------------------
Command-line options
-----------------------------------------------------------------------------
'''


def entry_point():
    PloterrsTool().main()


if __name__ == '__main__':
    entry_point()



from ._version import get_versions

from .core.segment import Segment
from .core.systems import WESTSystem
from .core.states import BasisState, TargetState
from .core import _rc

__all__ = ['Segment', 'WESTSystem', 'BasisState', 'TargetState', '_rc']

rc = _rc.WESTRC()

__version__ = get_versions()["version"]

del get_versions

from . import _version

__version__ = _version.get_versions()['version']



import logging
import sys

from warnings import warn
from westpa.tools import WESTTool, WESTDataReader
from westpa.core.segment import Segment

log = logging.getLogger('w_dumpsegs')


class WDumpSegs(WESTTool):
    prog = 'w_dumpsegs'
    description = '''\
Dump segment data as text. This is very inefficient, so this tool should be used
as a last resort (use hdfview/h5ls to look at data, and access HDF5 directly for
significant analysis tasks).
'''

    def __init__(self):
        super().__init__()
        self.data_reader = WESTDataReader()
        self.n_iter = None
        self.output_file = None
        self.print_pcoords = False

    def add_args(self, parser):
        self.data_reader.add_args(parser)

        parser.add_argument(
            '-p',
            '--print-pcoords',
            dest='print_pcoords',
            action='store_true',
            help='print initial and final progress coordinates for each segment',
        )
        parser.add_argument(
            '-i', '--iteration', dest='n_iter', type=int, help='Use data from iteration N_ITER (default: last complete iteration)'
        )
        parser.add_argument(
            '-o', '--output', dest='output_file', help='Store output in OUTPUT_FILE (default: write to standard output).'
        )

    def process_args(self, args):
        self.data_reader.process_args(args)
        self.data_reader.open()
        self.n_iter = args.n_iter or self.data_reader.current_iteration - 1
        self.output_file = open(args.output_file, 'wt') if args.output_file else sys.stdout
        self.print_pcoords = args.print_pcoords

    def go(self):
        segments = self.data_reader.get_segments(self.n_iter)

        max_seg_id_len = len(str(max(segment.seg_id for segment in segments)))
        max_status_name_len = max(list(map(len, iter(Segment.status_names.values()))))
        max_endpoint_type_len = max(list(map(len, iter(Segment.endpoint_type_names.values()))))
        max_n_parents_len = len(str(max(len(segment.wtg_parent_ids) for segment in segments)))

        report_line = (
            '{segment.n_iter:d}  {segment.seg_id:{max_seg_id_len}d}  {segment.weight:20.14g}'
            + '  {status_name:{max_status_name_len}s} ({segment.status})'
            + '  {segment.walltime:<12.6g} {segment.cputime:<12.6g}'
            + '  {endpoint_type_name:{max_endpoint_type_len}s} ({segment.endpoint_type})'
            + '  {n_parents:{max_n_parents_len}d} {segment.parent_id:{max_seg_id_len}d} {parents_str}'
            + '\n'
        )
        pcoord_lines = '  pcoord[0]  = {init_pcoord}\n  pcoord[-1] = {final_pcoord}' + '\n'
        for (_seg_id, segment) in enumerate(segments):
            parents_str = '[' + ', '.join(map(str, sorted(segment.wtg_parent_ids))) + ']'
            init_pcoord_str = '[' + ', '.join('{pcval:<12.6g}'.format(pcval=float(pce)) for pce in segment.pcoord[0]) + ']'
            final_pcoord_str = '[' + ', '.join('{pcval:<12.6g}'.format(pcval=float(pce)) for pce in segment.pcoord[-1]) + ']'
            self.output_file.write(
                report_line.format(
                    segment=segment,
                    status_name=segment.status_names[segment.status],
                    endpoint_type_name=segment.endpoint_type_names[segment.endpoint_type],
                    parents_str=parents_str,
                    n_parents=len(segment.wtg_parent_ids),
                    max_seg_id_len=max_seg_id_len,
                    max_status_name_len=max_status_name_len,
                    max_endpoint_type_len=max_endpoint_type_len,
                    max_n_parents_len=max_n_parents_len,
                )
            )
            if self.print_pcoords:
                self.output_file.write(pcoord_lines.format(init_pcoord=init_pcoord_str, final_pcoord=final_pcoord_str))


def entry_point():
    warn('w_dumpsegs is being deprecated.')
    WDumpSegs().main()


if __name__ == '__main__':
    entry_point()



import numpy as np
import numpy.ma as ma
import itertools


TOL = 1.0e-6


class UncertContainer:
    """Container to hold uncertainty measurements. Data is convert to np masked arrays
    to avoid possible numerical problems
    """

    def __init__(self, vals, vals_dmin, vals_dmax, mask=ma.nomask):
        super().__init__()

        # If input data already masked arrays extract unmasked data
        if ma.isMaskedArray(vals):
            vals = vals.data
        if ma.isMaskedArray(vals_dmin):
            vals_dmin = vals_dmin.data
        if ma.isMaskedArray(vals_dmax):
            vals_dmax = vals_dmax.data

        # Adjust negative values
        ineg = np.where(vals_dmin <= 0.0)
        vals_dmin[ineg] = TOL * vals[ineg]

        # Calculate weight based on fractional uncertainty
        diff = vals_dmax - vals_dmin
        diff_m = ma.masked_where(vals_dmax == vals_dmin, diff)

        self.vals = ma.masked_where(vals == 0.0, vals)

        self.wt = (self.vals / diff_m) ** 2
        self.uncert = diff_m / self.vals

        self.wt.fill_value = np.inf
        self.uncert.fill_vaule = np.inf

        assert np.all(self.wt.mask == self.uncert.mask)

        # Mask data if uncertainty is not finite or if any of the inputs were
        # already masked

        mm = ma.mask_or(self.wt.mask, mask)

        self.vals.mask = mm
        self.wt.mask = mm
        self.uncert.mask = mm
        self.dmin = ma.array(vals_dmin, mask=mm, fill_value=np.inf)
        self.dmax = ma.array(vals_dmax, mask=mm, fill_value=np.inf)

        self.mask = ma.getmaskarray(self.vals)

    def __getitem__(self, indx):
        vals = self.vals[indx]
        dmin = self.dmin[indx]
        dmax = self.dmax[indx]

        if isinstance(vals, ma.core.MaskedConstant):
            dum = np.zeros((1,))
            return UncertContainer(dum.copy(), dum.copy(), dum.copy())
        elif isinstance(vals, (float, int, np.float, np.int)):
            return UncertContainer(np.array([vals]), np.array([dmin]), np.array([dmax]))
        elif isinstance(vals, np.ndarray):
            return UncertContainer(vals, dmin, dmax, mask=vals.mask)
        else:
            raise TypeError

    def __setitem__(self, indx, value):
        if isinstance(value, UncertContainer):
            self.vals[indx] = value.vals
            self.dmin[indx] = value.dmin
            self.dmax[indx] = value.dmax
            self.wt[indx] = value.wt
            self.uncert[indx] = value.uncert
        else:
            raise TypeError('Can only set values with an UncertContainer object')

    def __repr__(self):
        return 'shape={} vals={} dmin={} dmax={} vals.mask={}'.format(
            self.vals.shape, self.vals, self.dmin, self.dmax, self.vals.mask
        )

    def __add__(self, value):
        if isinstance(value, UncertContainer):
            vals = self.vals + value.vals
            dmin = self.dmin + value.dmin
            dmax = self.dmax + value.dmax

            return UncertContainer(vals, dmin, dmax, mask=vals.mask)
        elif isinstance(value, (float, int, np.float, np.int)):
            vals = self.vals + value
            dmin = self.dmin + value
            dmax = self.dmax + value

            return UncertContainer(vals, dmin, dmax, mask=vals.mask)
        else:
            raise TypeError('Attempt to add value of unsupported type')

    def __sub__(self, value):
        if isinstance(value, UncertContainer):
            vals = self.vals - value.vals
            dmin = self.dmin - value.dmin
            dmax = self.dmax - value.dmax

            return UncertContainer(vals, dmin, dmax, mask=vals.mask)
        else:
            raise TypeError('Attempted to subtract by value of unsupported type')

    def __mul__(self, value):
        if isinstance(value, UncertContainer):
            vals = self.vals * value.vals
            dmin = self.dmin * value.dmin
            dmax = self.dmax * value.dmax

            return UncertContainer(vals, dmin, dmax, mask=vals.mask)

        elif isinstance(value, (float, int, np.float, np.int)):
            vals = self.vals * value
            dmin = self.dmin * value
            dmax = self.dmax * value

            return UncertContainer(vals, dmin, dmax, mask=vals.mask)
        else:
            raise TypeError('Attempted to multiply by value of unsupported type')

    def __truediv__(self, value):
        if isinstance(value, UncertContainer):
            vals = self.vals / value.vals
            dmin = self.dmin / value.dmax
            dmax = self.dmax / value.dmin

            return UncertContainer(vals, dmin, dmax, mask=vals.mask)
        else:
            raise TypeError('Attempted to divide by unsupported type')

    def transpose(self):
        vals = self.vals.T
        dmin = self.dmin.T
        dmax = self.dmax.T

        return UncertContainer(vals, dmin, dmax, mask=vals.mask)

    def recip(self):
        vals = 1.0 / self.vals
        dmin = 1.0 / self.dmax
        dmax = 1.0 / self.dmin

        return UncertContainer(vals, dmin, dmax, mask=vals.mask)

    def update_mask(self):
        self.vals.mask = self.mask
        self.dmin.mask = self.mask
        self.dmax.mask = self.mask
        self.wt.mask = self.mask
        self.uncert.mask = self.mask

    def concatenate(self, value, axis=0):
        """Concatentate UncertContainer value to self.
        Assumes that if dimensions of self and value do not match, to
        add a np.newaxis along axis of value
        """

        if isinstance(value, UncertContainer):
            if value.vals.ndim == self.vals.ndim:
                vals = value.vals
                dmin = value.dmin
                dmax = value.dmax
                wt = value.wt
                uncert = value.uncert
                mask = value.mask
            elif (value.vals.ndim + 1) == self.vals.ndim:
                vals = ma.expand_dims(value.vals, axis)
                dmin = ma.expand_dims(value.dmin, axis)
                dmax = ma.expand_dims(value.dmax, axis)
                wt = ma.expand_dims(value.wt, axis)
                uncert = ma.expand_dims(value.uncert, axis)
                mask = np.expand_dims(value.mask, axis)
            else:
                raise ValueError('Could not propery match dimensionality')

            self.vals = ma.concatenate((self.vals, vals), axis=axis)
            self.dmin = ma.concatenate((self.dmin, dmin), axis=axis)
            self.dmax = ma.concatenate((self.dmax, dmax), axis=axis)
            self.wt = ma.concatenate((self.wt, wt), axis=axis)
            self.uncert = ma.concatenate((self.uncert, uncert), axis=axis)

            self.mask = np.concatenate((self.mask, mask), axis=axis)
        else:
            raise ValueError('Can only concatenate with an UncertContainer object')

    def weighted_average(self, axis=0, expaxis=None):
        """Calculate weighted average of data along axis
        after optionally inserting a new dimension into the
        shape array at position expaxis
        """

        if expaxis is not None:
            vals = ma.expand_dims(self.vals, expaxis)
            dmin = ma.expand_dims(self.dmin, expaxis)
            dmax = ma.expand_dims(self.dmax, expaxis)
            wt = ma.expand_dims(self.wt, expaxis)
        else:
            vals = self.vals
            wt = self.wt
            dmin = self.dmin
            dmax = self.dmax

        # Get average value
        avg, norm = ma.average(vals, axis=axis, weights=wt, returned=True)
        avg_ex = ma.expand_dims(avg, 0)

        # Calculate weighted uncertainty
        wtmax = ma.max(wt, axis=axis)
        neff = norm / wtmax  # Effective number of samples based on uncertainties

        # Seeking max deviation from the average; if above avg use max, if below use min
        term = np.empty_like(vals)

        indices = np.where(vals > avg_ex)
        i0 = indices[0]
        irest = indices[1:]
        ii = tuple(x for x in itertools.chain([i0], irest))
        jj = tuple(x for x in itertools.chain([np.zeros_like(i0)], irest))
        term[ii] = (dmax[ii] - avg_ex[jj]) ** 2

        indices = np.where(vals <= avg_ex)
        i0 = indices[0]
        irest = indices[1:]
        ii = tuple(x for x in itertools.chain([i0], irest))
        jj = tuple(x for x in itertools.chain([np.zeros_like(i0)], irest))
        term[ii] = (avg_ex[jj] - dmin[ii]) ** 2

        dsum = ma.sum(term * wt, axis=0)  # Sum for weighted average of deviations

        dev = 0.5 * np.sqrt(dsum / (norm * neff))

        if isinstance(avg, (float, np.float)):
            avg = avg_ex

        tmp_min = avg - dev
        ii = np.where(tmp_min < 0)
        tmp_min[ii] = TOL * avg[ii]

        return UncertContainer(avg, tmp_min, avg + dev)



class ArgumentError(RuntimeError):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class AnalysisMixin:
    def __init__(self):
        super().__init__()

    def add_args(self, parser, upcall=True):
        if upcall:
            try:
                upfunc = super().add_args
            except AttributeError:
                pass
            else:
                upfunc(parser)

    def process_args(self, args, upcall=True):
        if upcall:
            try:
                upfunc = super().process_args
            except AttributeError:
                pass
            else:
                upfunc(args)



'''Miscellaneous support functions for WEST and WEST tools'''

import re


def parse_int_list(list_string):
    '''Parse a simple list consisting of integers or ranges of integers separated by commas. Ranges are specified
    as min:max, and include the maximum value (unlike Python's ``range``).  Duplicate values are ignored.
    Returns the result as a sorted list.  Raises ValueError if the list cannot be parsed.'''

    try:
        entries = set()
        fields = re.split(r'\s*[;,]\s*', list_string)
        for field in fields:
            if ':' in field:
                lb, ub = list(map(int, re.split(r'\s*:\s*', field)))
                entries.update(list(range(lb, ub + 1)))
            else:
                entries.add(int(field))
    except (ValueError, TypeError):
        raise ValueError('invalid integer range string {!r}'.format(list_string))
    else:
        return sorted(entries)



import logging
import multiprocessing
import os
import random
import signal
import sys
import threading
import traceback

import westpa.work_managers as work_managers
from .core import WorkManager, WMFuture

log = logging.getLogger(__name__)

# Tasks are tuples ('task', task_id, fn, args, kwargs).
# Results are tuples (rtype, task_id, payload) where rtype is 'result' or 'exception' and payload is the return value
# or exception, respectively.

task_shutdown_sentinel = ('shutdown', None, None, (), {})
result_shutdown_sentinel = ('shutdown', None, None)


class ProcessWorkManager(WorkManager):
    '''A work manager using the ``multiprocessing`` module.

    Notes
    -----

    On MacOS, as of Python 3.8 the default start method for multiprocessing launching new processes was changed from fork to spawn.
    In general, spawn is more robust and efficient, however it requires serializability of everything being passed to the child process.
    In contrast, fork is much less memory efficient, as it makes a full copy of everything in the parent process.
    However, it does not require picklability.

    So, on MacOS, the method for launching new processes is explicitly changed to fork from the (MacOS-specific) default of spawn.
    Unix should default to fork.

    See https://docs.python.org/3/library/multiprocessing.html#contexts-and-start-methods and
    https://docs.python.org/3/library/multiprocessing.html#the-spawn-and-forkserver-start-methods for more details.
    '''

    @classmethod
    def from_environ(cls, wmenv=None):
        if wmenv is None:
            wmenv = work_managers.environment.default_env
        return cls(wmenv.get_val('n_workers', multiprocessing.cpu_count(), int))

    def __init__(self, n_workers=None, shutdown_timeout=1):
        super().__init__()

        try:
            if sys.platform == 'darwin':  # MacOS
                multiprocessing.set_start_method('fork')
                log.debug('setting multiprocessing start method to fork')
        except RuntimeError:
            log.debug('failed to set start method to fork')

        self.n_workers = n_workers or multiprocessing.cpu_count()
        self.workers = None
        self.task_queue = multiprocessing.Queue()
        self.result_queue = multiprocessing.Queue()
        self.receive_thread = None
        self.pending = None

        self.shutdown_received = False
        self.shutdown_timeout = shutdown_timeout or 1

    def task_loop(self):
        # Close standard input, so we don't get SIGINT from ^C
        try:
            sys.stdin.close()
        except Exception as e:
            log.info("can't close stdin: {}".format(e))

        # (re)initialize random number generator in this process
        random.seed()

        while not self.shutdown_received:
            message, task_id, fn, args, kwargs = self.task_queue.get()[:5]

            if message == 'shutdown':
                break

            try:
                result = fn(*args, **kwargs)
            except BaseException as e:
                result_tuple = ('exception', task_id, (e, traceback.format_exc()))
            else:
                result_tuple = ('result', task_id, result)
            self.result_queue.put(result_tuple)

        log.debug('exiting task_loop')
        return

    def results_loop(self):
        while not self.shutdown_received:
            message, task_id, payload = self.result_queue.get()[:3]

            if message == 'shutdown':
                break
            elif message == 'exception':
                future = self.pending.pop(task_id)
                future._set_exception(*payload)
            elif message == 'result':
                future = self.pending.pop(task_id)
                future._set_result(payload)
            else:
                raise AssertionError('unknown message {!r}'.format((message, task_id, payload)))

        log.debug('exiting results_loop')

    def submit(self, fn, args=None, kwargs=None):
        ft = WMFuture()
        log.debug('dispatching {!r}'.format(fn))
        self.pending[ft.task_id] = ft
        self.task_queue.put(('task', ft.task_id, fn, args or (), kwargs or {}))
        return ft

    def startup(self):
        from . import environment

        if not self.running:
            log.debug('starting up work manager {!r}'.format(self))
            self.running = True
            self.workers = [
                multiprocessing.Process(target=self.task_loop, name='worker-{:d}-{:x}'.format(i, id(self)))
                for i in range(self.n_workers)
            ]

            pi_name = '{}_PROCESS_INDEX'.format(environment.WMEnvironment.env_prefix)
            for iworker, worker in enumerate(self.workers):
                os.environ[pi_name] = str(iworker)
                worker.start()
            try:
                del os.environ[pi_name]
            except KeyError:
                pass

            self.pending = dict()

            self.receive_thread = threading.Thread(target=self.results_loop, name='receiver')
            self.receive_thread.daemon = True
            self.receive_thread.start()

    def _empty_queues(self):
        while not self.task_queue.empty():
            try:
                self.task_queue.get(block=False)
            except multiprocessing.queues.Empty:
                break

        while not self.result_queue.empty():
            try:
                self.result_queue.get(block=False)
            except multiprocessing.queues.Empty:
                break

    def shutdown(self):
        if self.running:
            log.debug('shutting down {!r}'.format(self))
            self._empty_queues()

            # Send shutdown signal
            for _i in range(self.n_workers):
                self.task_queue.put(task_shutdown_sentinel, block=False)

            for worker in self.workers:
                worker.join(self.shutdown_timeout)
                if worker.is_alive():
                    log.debug('sending SIGINT to worker process {:d}'.format(worker.pid))
                    os.kill(worker.pid, signal.SIGINT)
                    worker.join(self.shutdown_timeout)
                    if worker.is_alive():
                        log.warning('sending SIGKILL to worker process {:d}'.format(worker.pid))
                        os.kill(worker.pid, signal.SIGKILL)
                        worker.join()

                    log.debug('worker process {:d} terminated with code {:d}'.format(worker.pid, worker.exitcode))
                else:
                    log.debug('worker process {:d} terminated gracefully with code {:d}'.format(worker.pid, worker.exitcode))

            self._empty_queues()
            self.result_queue.put(result_shutdown_sentinel)
            self.running = False



import numpy as np
from westpa.core.binning import FuncBinMapper
import logging
import westpa

log = logging.getLogger(__name__)


def map_mab(coords, mask, output, *args, **kwargs):
    '''Binning which adaptively places bins based on the positions of extrema segments and
    bottleneck segments, which are where the difference in probability is the greatest
    along the progress coordinate. Operates per dimension and places a fixed number of
    evenly spaced bins between the segments with the min and max pcoord values. Extrema and
    bottleneck segments are assigned their own bins.'''

    pca = kwargs.get("pca")
    bottleneck = kwargs.get("bottleneck")
    direction = kwargs.get("direction")
    skip = kwargs.get("skip")
    nbins_per_dim = kwargs.get("nbins_per_dim")
    mab_log = kwargs.get("mab_log")
    ndim = len(nbins_per_dim)

    if not np.any(mask):
        return output

    allcoords = np.copy(coords)
    allmask = np.copy(mask)

    weights = None
    isfinal = None
    splitting = False
    report = False

    # the segments should be sent in by the driver as half initial segments and half final segments
    # allcoords contains all segments
    # coords should contain ONLY final segments
    if coords.shape[1] > ndim:
        if coords[0, -1] == 0:
            report = True
        if coords.shape[1] > ndim + 1:
            isfinal = allcoords[:, ndim + 1].astype(np.bool_)
        else:
            isfinal = np.ones(coords.shape[0], dtype=np.bool_)
        coords = coords[isfinal, :ndim]
        weights = allcoords[isfinal, ndim + 0]
        mask = mask[isfinal]
        splitting = True

    if not np.any(mask):
        coords = allcoords[:, :ndim]
        mask = allmask
        weights = None
        splitting = False

    varcoords = np.copy(coords)
    originalcoords = np.copy(coords)
    if pca and len(output) > 1:
        colavg = np.mean(coords, axis=0)
        for i in range(len(coords)):
            for j in range(len(coords[i])):
                varcoords[i][j] = coords[i][j] - colavg[j]
        covcoords = np.cov(np.transpose(varcoords), aweights=weights)
        eigval, eigvec = np.linalg.eigh(covcoords)
        eigvec = eigvec[:, np.argmax(np.absolute(eigvec), axis=1)]
        for i in range(len(eigvec)):
            if eigvec[i, i] < 0:
                eigvec[:, i] = -1 * eigvec[:, i]
        for i in range(ndim):
            for j in range(len(output)):
                coords[j][i] = np.dot(varcoords[j], eigvec[:, i])

    maxlist = []
    minlist = []
    difflist = []
    flipdifflist = []
    for n in range(ndim):
        # identify the boundary segments
        maxcoord = np.max(coords[mask, n])
        mincoord = np.min(coords[mask, n])
        maxlist.append(maxcoord)
        minlist.append(mincoord)

        # detect the bottleneck segments, this uses the weights
        if splitting:
            temp = np.column_stack((originalcoords[mask, n], weights[mask]))
            sorted_indices = temp[:, 0].argsort()
            temp = temp[sorted_indices]
            for p in range(len(temp)):
                if temp[p][1] == 0:
                    temp[p][1] = 10**-323
            fliptemp = np.flipud(temp)

            difflist.append(None)
            flipdifflist.append(None)
            maxdiff = 0
            flipmaxdiff = 0
            for i in range(1, len(temp) - 1):
                comprob = 0
                flipcomprob = 0
                j = i + 1
                while j < len(temp):
                    comprob = comprob + temp[j][1]
                    flipcomprob = flipcomprob + fliptemp[j][1]
                    j = j + 1
                diff = -np.log(comprob) + np.log(temp[i][1])
                if diff > maxdiff:
                    difflist[n] = temp[i][0]
                    maxdiff = diff
                flipdiff = -np.log(flipcomprob) + np.log(fliptemp[i][1])
                if flipdiff > flipmaxdiff:
                    flipdifflist[n] = fliptemp[i][0]
                    flipmaxdiff = flipdiff

    if mab_log and report:
        westpa.rc.pstatus("################ MAB stats ################")
        westpa.rc.pstatus("minima in each dimension:      {}".format(minlist))
        westpa.rc.pstatus("maxima in each dimension:      {}".format(maxlist))
        westpa.rc.pstatus("direction in each dimension:   {}".format(direction))
        westpa.rc.pstatus("skip in each dimension:        {}".format(skip))
        westpa.rc.pstatus("###########################################")
        westpa.rc.pflush()

    # assign segments to bins
    # the total number of linear bins is the boundary base
    boundary_base = np.prod(nbins_per_dim)

    # the bottleneck base is offset by the number of boundary walkers,
    # which is two per dimension unless there is a direction specified
    # in a particluar dimension, then it's just one
    bottleneck_base = boundary_base

    for i in range(0, ndim):
        if direction[i] != 0:
            bottleneck_base += 1
        else:
            bottleneck_base += 2

    # if a dimension is being "skipped", leave only one bin total as
    # the offset
    for i in range(0, ndim):
        if skip[i] != 0:
            boundary_base -= nbins_per_dim[i] - 1

    for i in range(len(output)):
        if not allmask[i]:
            continue

        # special means either a boundary or bottleneck walker (not a walker in the linear space)
        special = False
        # this holder is the bin number, which only needs to be unique for different walker groups
        holder = 0
        if splitting:
            for n in range(ndim):
                coord = allcoords[i][n]

                # if skipped, just assign the walkers to the same bin (offset of boundary base)
                if skip[n] != 0:
                    holder = boundary_base + n
                    break

                # assign bottlenecks, taking directionality into account
                if bottleneck:
                    if direction[n] < 0:
                        if coord == flipdifflist[n]:
                            holder = bottleneck_base + n
                            special = True
                            break

                    if direction[n] > 0:
                        if coord == difflist[n]:
                            holder = bottleneck_base + n
                            special = True
                            break

                    if direction[n] == 0:
                        if coord == difflist[n]:
                            holder = bottleneck_base + n
                            special = True
                            break
                        elif coord == flipdifflist[n]:
                            holder = bottleneck_base + n + 1
                            special = True
                            break

                # assign boundary walkers, taking directionality into account
                if direction[n] < 0:
                    if coord == minlist[n]:
                        holder = boundary_base + n
                        special = True
                        break

                elif direction[n] > 0:
                    if coord == maxlist[n]:
                        holder = boundary_base + n
                        special = True
                        break

                elif direction[n] == 0:
                    if coord == minlist[n]:
                        holder = boundary_base + n
                        special = True
                        break
                    elif coord == maxlist[n]:
                        holder = boundary_base + n + 1
                        special = True
                        break

        # the following are for the "linear" portion
        if not special:
            for n in range(ndim):

                # if skipped, it's added to the same bin as the special walkers above
                if skip[n] != 0:
                    holder = boundary_base + n
                    break

                coord = allcoords[i][n]
                nbins = nbins_per_dim[n]
                minp = minlist[n]
                maxp = maxlist[n]

                bins = np.linspace(minp, maxp, nbins + 1)
                bin_number = np.digitize(coord, bins) - 1

                if isfinal is None or not isfinal[i]:
                    if bin_number >= nbins:
                        bin_number = nbins - 1
                    elif bin_number < 0:
                        bin_number = 0
                elif bin_number >= nbins or bin_number < 0:
                    if np.isclose(bins[-1], coord):
                        bin_number = nbins - 1
                    elif np.isclose(bins[0], coord):
                        bin_number = 0
                    else:
                        raise ValueError("Walker out of boundary")

                holder += bin_number * np.prod(nbins_per_dim[:n])

        # output is the main list that, for each segment, holds the bin assignment
        output[i] = holder

    return output


class MABBinMapper(FuncBinMapper):
    '''Adaptively place bins in between minimum and maximum segments along
    the progress coordinte. Extrema and bottleneck segments are assigned
    to their own bins.'''

    def __init__(self, nbins, direction=None, skip=None, bottleneck=True, pca=False, mab_log=False):
        # Verifying parameters
        if nbins is None:
            raise ValueError("nbins_per_dim is missing")
        ndim = len(nbins)

        if direction is None:
            direction = [0] * ndim
        elif len(direction) != ndim:
            direction = [0] * ndim
            log.warn("Direction list is not the correct dimensions, setting to defaults.")

        if skip is None:
            skip = [0] * ndim
        elif len(skip) != ndim:
            skip = [0] * ndim
            log.warn("Skip list is not the correct dimensions, setting to defaults.")

        kwargs = dict(nbins_per_dim=nbins, direction=direction, skip=skip, bottleneck=bottleneck, pca=pca, mab_log=mab_log)
        # the following is neccessary because functional bin mappers need to "reserve"
        # bins and tell the sim manager how many bins they will need to use, this is
        # determined by taking all direction/skipping info into account
        n_total_bins = np.prod(nbins)
        for i in range(0, ndim):
            if skip[i] == 0:
                if direction[i] != 0:
                    n_total_bins += 1 + 1 * bottleneck
                else:
                    n_total_bins += 2 + 2 * bottleneck
            else:
                n_total_bins -= nbins[i] - 1
                n_total_bins += 1 * ndim  # or else it will be one bin short
        super().__init__(map_mab, n_total_bins, kwargs=kwargs)



import concurrent.futures as cf
import functools
import inspect
import operator
import os

import mdtraj
import numpy as np

from tqdm import tqdm
from typing import Callable
from westpa.analysis.core import Walker, Trace
from westpa.core.states import InitialState
from westpa.core.h5io import WESTIterationFile


class Trajectory:
    """A callable that returns the trajectory of a walker or trace.

    Parameters
    ----------
    fget : callable
        Function for retrieving a single trajectory segment. Must take a
        :class:`Walker` instance as its first argument and accept a boolean
        keyword argument `include_initpoint`. The function should return a
        sequence (e.g., a list or ndarray) representing the trajectory of
        the walker. If `include_initpoint` is True, the trajectory segment
        should include its initial point. Otherwise, the trajectory segment
        should exclude its initial point.
    fconcat : callable, optional
        Function for concatenating trajectory segments. Must take a sequence
        of trajectory segments as input and return their concatenation. The
        default concatenation function is :func:`concatenate`.

    """

    def __init__(self, fget=None, *, fconcat=None):
        if fget is None:
            return functools.partial(self.__init__, fconcat=fconcat)

        if 'include_initpoint' not in inspect.signature(fget).parameters:
            raise ValueError("'fget' must accept a parameter 'include_initpoint'")

        self._fget = fget
        self.fconcat = fconcat

        self._segment_collector = SegmentCollector(self)

    @property
    def segment_collector(self):
        """SegmentCollector: Segment retrieval manager."""
        return self._segment_collector

    @property
    def fget(self):
        """callable: Function for getting trajectory segments."""
        return self._fget

    @property
    def fconcat(self):
        """callable: Function for concatenating trajectory segments."""
        return self._fconcat

    @fconcat.setter
    def fconcat(self, value):
        if value is None:
            value = concatenate
        elif not isinstance(value, Callable):
            raise TypeError("'fconcat' must be a callable object")
        self._fconcat = value

    def __call__(self, obj, include_initpoint=True, **kwargs):
        if isinstance(obj, Walker):
            value = self.fget(obj, include_initpoint=include_initpoint, **kwargs)
            self._validate_segment(value)
            return value
        if isinstance(obj, Trace):
            initpoint_mask = np.full(len(obj), False)
            initpoint_mask[0] = include_initpoint
            segments = self.segment_collector.get_segments(obj, initpoint_mask, **kwargs)
            return self.fconcat(segments)
        raise TypeError('argument must be a Walker or Trace instance')

    def _validate_segment(self, value):
        if not hasattr(value, '__getitem__'):
            msg = f"{type(value).__name__!r} object can't be concatenated"
            raise TypeError(msg)


class SegmentCollector:
    """An object that manages the retrieval of trajectory segments.

    Parameters
    ----------
    trajectory : Trajectory
        The trajectory to which the segment collector is attached.
    use_threads : bool, default False
        Whether to use a pool of threads to retrieve trajectory segments
        asynchronously. Setting this parameter to True may be may be
        useful when segment retrieval is an I/O bound task.
    max_workers : int, optional
        Maximum number of threads to use. The default value is specified in the
        `ThreadPoolExecutor <https://docs.python.org/3/library/concurrent.futures.html#concurrent.futures.ThreadPoolExecutor>`_
        documentation.
    show_progress : bool, default False
        Whether to show a progress bar when retrieving multiple segments.

    """

    def __init__(self, trajectory, use_threads=False, max_workers=None, show_progress=False):
        self.trajectory = trajectory
        self.use_threads = use_threads
        self.max_workers = max_workers
        self.show_progress = show_progress

    @property
    def trajectory(self):
        return self._trajectory

    @trajectory.setter
    def trajectory(self, value):
        if not isinstance(value, Trajectory):
            msg = f'trajectory must be an instance of {Trajectory}'
            raise TypeError(msg)
        self._trajectory = value

    @property
    def use_threads(self):
        return self._use_threads

    @use_threads.setter
    def use_threads(self, value):
        if not isinstance(value, bool):
            raise TypeError('use_threads must be True or False')
        self._use_threads = value

    @property
    def max_workers(self):
        return self._max_workers

    @max_workers.setter
    def max_workers(self, value):
        if value is None:
            self._max_workers = None
            return
        if value <= 0:
            raise ValueError('max_workers must be greater than 0')
        self._max_workers = value

    @property
    def show_progress(self):
        return self._show_progress

    @show_progress.setter
    def show_progress(self, value):
        if not isinstance(value, bool):
            raise ValueError('show_progress must be True or False')
        self._show_progress = value

    def get_segments(self, walkers, initpoint_mask=None, **kwargs):
        """Retrieve the trajectories of multiple walkers.

        Parameters
        ----------
        walkers : sequence of Walker
            The walkers for which to retrieve trajectories.
        initpoint_mask : sequence of bool, optional
            A Boolean mask indicating whether each trajectory segment should
            include (True) or exclude (False) its initial point. Default is
            all True.

        Returns
        -------
        list of sequences
            The trajectory of each walker.

        """
        if initpoint_mask is None:
            initpoint_mask = np.full(len(walkers), True)
        else:
            initpoint_mask = np.asarray(initpoint_mask, dtype=bool)

        get_segment = functools.partial(self.trajectory, **kwargs)

        tqdm_kwargs = dict(
            desc='Retrieving segments',
            disable=(not self.show_progress),
            position=0,
            total=len(walkers),
        )

        if self.use_threads:
            with cf.ThreadPoolExecutor(self.max_workers) as executor:
                future_to_key = {
                    executor.submit(get_segment, walker, include_initpoint=i): key
                    for key, (walker, i) in enumerate(zip(walkers, initpoint_mask))
                }
                futures = list(tqdm(cf.as_completed(future_to_key), **tqdm_kwargs))
                futures.sort(key=future_to_key.get)
                segments = (future.result() for future in futures)
        else:
            it = (get_segment(walker, include_initpoint=i) for walker, i in zip(walkers, initpoint_mask))
            segments = tqdm(it, **tqdm_kwargs)

        return list(segments)


class BasicMDTrajectory(Trajectory):
    """Trajectory reader for MD trajectories stored as in the
    `Basic Tutorial <https://github.com/westpa/westpa_tutorials/tree/main/basic_nacl>`_.

    Parameters
    ----------
    top : str or mdtraj.Topology, default 'bstate.pdb'
    traj_ext : str, default '.dcd'
    state_ext : str, default '.xml'
    sim_root : str, default '.'

    """

    def __init__(self, top='bstate.pdb', traj_ext='.dcd', state_ext='.xml', sim_root='.'):
        self.top = top
        self.traj_ext = traj_ext
        self.state_ext = state_ext
        self.sim_root = sim_root

        def fget(walker, include_initpoint=True, atom_indices=None, sim_root=None):
            sim_root = sim_root or self.sim_root

            if isinstance(self.top, str):
                top = os.path.join(sim_root, 'common_files', self.top)
            else:
                top = self.top

            path = os.path.join(
                sim_root,
                'traj_segs',
                format(walker.iteration.number, '06d'),
                format(walker.index, '06d'),
                'seg' + self.traj_ext,
            )
            if top is not None:
                traj = mdtraj.load(path, top=top)
            else:
                traj = mdtraj.load(path)

            if include_initpoint:
                parent = walker.parent

                if isinstance(parent, InitialState):
                    if parent.istate_type == InitialState.ISTATE_TYPE_BASIS:
                        path = os.path.join(
                            sim_root,
                            'bstates',
                            parent.basis_state.auxref,
                        )
                    else:
                        path = os.path.join(
                            sim_root,
                            'istates',
                            str(parent.iter_created),
                            str(parent.state_id) + self.state_ext,
                        )
                else:
                    path = os.path.join(
                        sim_root,
                        'traj_segs',
                        format(walker.iteration.number - 1, '06d'),
                        format(parent.index, '06d'),
                        'seg' + self.state_ext,
                    )

                frame = mdtraj.load(path, top=traj.top)
                traj = frame.join(traj, check_topology=False)

            if atom_indices is not None:
                traj.atom_slice(atom_indices, inplace=True)

            return traj

        super().__init__(fget)

        self.segment_collector.use_threads = True
        self.segment_collector.show_progress = True


class HDF5MDTrajectory(Trajectory):
    """Trajectory reader for MD trajectories stored by the HDF5 framework."""

    def __init__(self):
        def fget(walker, include_initpoint=True, atom_indices=None):
            iteration = walker.iteration

            try:
                link = iteration.h5group['trajectories']
            except KeyError:
                msg = 'the HDF5 framework does not appear to have been used to store trajectories for this run'
                raise ValueError(msg)

            with WESTIterationFile(link.file.filename) as traj_file:
                traj = traj_file.read_as_traj(
                    iteration=iteration.number,
                    segment=walker.index,
                    atom_indices=atom_indices,
                )

            if include_initpoint:
                parent = walker.parent

                if isinstance(parent, InitialState):
                    if parent.istate_type == InitialState.ISTATE_TYPE_BASIS:
                        link = walker.run.h5file.get_iter_group(0)['trajectories']
                        with WESTIterationFile(link.file.filename) as traj_file:
                            frame = traj_file.read_as_traj(
                                iteration=0,
                                segment=parent.basis_state_id,
                                atom_indices=atom_indices,
                            )
                    elif parent.istate_type == InitialState.ISTATE_TYPE_GENERATED:
                        link = walker.run.h5file.get_iter_group(0)['trajectories']
                        istate_iter = -int(parent.iter_created)  # the conversion to int is because iter_created is uint
                        with WESTIterationFile(link.file.filename) as traj_file:
                            frame = traj_file.read_as_traj(
                                iteration=istate_iter,
                                segment=parent.state_id,
                                atom_indices=atom_indices,
                            )
                    else:
                        raise ValueError('unsupported initial state type: %d' % parent.istate_type)
                else:
                    frame = fget(parent, include_initpoint=False, atom_indices=atom_indices)[-1]
                traj = frame.join(traj, check_topology=False)

            return traj

        super().__init__(fget)

        self.segment_collector.use_threads = False
        self.segment_collector.show_progress = True


def concatenate(segments):
    """Return the concatenation of a sequence of trajectory segments.

    Parameters
    ----------
    segments : sequence of sequences
        A sequence of trajectory segments.

    Returns
    -------
    sequence
        The concatenation of `segments`.

    """
    if isinstance(segments[0], np.ndarray):
        return np.concatenate(segments)
    if isinstance(segments[0], mdtraj.Trajectory):
        return segments[0].join(segments[1:], check_topology=False)
    return functools.reduce(operator.concat, segments)



import warnings
import numpy


def load_npy_or_text(filename):
    '''Load an array from an existing .npy file, or read a text file and
    convert to a NumPy array.  In either case, return a NumPy array.  If a
    pickled NumPy dataset is found, memory-map it read-only.  If the specified
    file does not contain a pickled NumPy array, attempt to read the file using
    numpy.loadtxt(filename, **kwargs).'''

    f = open(filename, 'rb')
    try:
        f.seek(0)
    except IOError:
        # Not seekable - assume a text stream
        return numpy.loadtxt(filename)
    else:
        f.close()

    # File is seekable
    try:
        # try to mmap it
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            return numpy.load(filename, 'r')

    except IOError as e:
        if 'Failed to interpret' in str(e):
            pass
        else:
            raise

    return numpy.loadtxt(filename)



import logging

import numpy as np

from westpa.core.data_manager import weight_dtype
from westpa.tools import WESTMasterCommand, WESTParallelTool

from westpa.core import h5io
from westpa.core.kinetics import sequence_macro_flux_to_rate, WKinetics

from westpa.tools.kinetics_tool import WESTKineticsBase, AverageCommands

from westpa.mclib import mcbs_ci_correl, _1D_simple_eval_block, _2D_simple_eval_block


# From w_stateprobs
from westpa.core.binning import accumulate_state_populations_from_labeled

log = logging.getLogger('w_direct')


# This block is responsible for submitting a set of calculations to be bootstrapped over for a particular type of calculation.
# A property which wishes to be calculated should adhere to this format.
def _rate_eval_block(iblock, start, stop, nstates, data_input, name, mcbs_alpha, mcbs_nsets, mcbs_acalpha, do_correl, mcbs_enable):
    # Our rate estimator is a little more complex, so we've defined a custom evaluation block for it,
    # instead of just using the block evalutors that we've imported.
    results = []
    for istate in range(nstates):
        for jstate in range(nstates):
            if istate == jstate:
                continue
            kwargs = {'istate': istate, 'jstate': jstate}
            # Why are we sending in the total population dataset, instead of a sliced one?
            # It's a requirement of our estimator; we need to pull from any given i to j state in order to properly normalize
            # and avoid i to j rate constants which are affected by a third state k.
            # That is, we need the populations for both i and j, and it's easier to just send in the entire dataset.
            dataset = {'dataset': data_input['dataset'][:, istate, jstate], 'pops': data_input['pops']}
            ci_res = mcbs_ci_correl(
                dataset,
                estimator=sequence_macro_flux_to_rate,
                alpha=mcbs_alpha,
                n_sets=mcbs_nsets,
                autocorrel_alpha=mcbs_acalpha,
                subsample=np.mean,
                do_correl=do_correl,
                mcbs_enable=mcbs_enable,
                estimator_kwargs=kwargs,
            )
            results.append((name, iblock, istate, jstate, (start, stop) + ci_res))

    return results


# The old w_kinetics
class DKinetics(WESTKineticsBase, WKinetics):
    subcommand = 'init'
    default_kinetics_file = 'direct.h5'
    default_output_file = 'direct.h5'
    help_text = 'calculate state-to-state kinetics by tracing trajectories'
    description = '''\
Calculate state-to-state rates and transition event durations by tracing
trajectories.

A bin assignment file (usually "assign.h5") including trajectory labeling
is required (see "w_assign --help" for information on generating this file).

This subcommand for w_direct is used as input for all other w_direct
subcommands, which will convert the flux data in the output file into
average rates/fluxes/populations with confidence intervals.

-----------------------------------------------------------------------------
Output format
-----------------------------------------------------------------------------

The output file (-o/--output, by default "direct.h5") contains the
following datasets:

  ``/conditional_fluxes`` [iteration][state][state]
    *(Floating-point)* Macrostate-to-macrostate fluxes. These are **not**
    normalized by the population of the initial macrostate.

  ``/conditional_arrivals`` [iteration][stateA][stateB]
    *(Integer)* Number of trajectories arriving at state *stateB* in a given
    iteration, given that they departed from *stateA*.

  ``/total_fluxes`` [iteration][state]
    *(Floating-point)* Total flux into a given macrostate.

  ``/arrivals`` [iteration][state]
    *(Integer)* Number of trajectories arriving at a given state in a given
    iteration, regardless of where they originated.

  ``/duration_count`` [iteration]
    *(Integer)* The number of event durations recorded in each iteration.

  ``/durations`` [iteration][event duration]
    *(Structured -- see below)*  Event durations for transition events ending
    during a given iteration. These are stored as follows:

      istate
        *(Integer)* Initial state of transition event.
      fstate
        *(Integer)* Final state of transition event.
      duration
        *(Floating-point)* Duration of transition, in units of tau.
      weight
        *(Floating-point)* Weight of trajectory at end of transition, **not**
        normalized by initial state population.

Because state-to-state fluxes stored in this file are not normalized by
initial macrostate population, they cannot be used as rates without further
processing. The ``w_direct kinetics`` command is used to perform this normalization
while taking statistical fluctuation and correlation into account. See
``w_direct kinetics --help`` for more information.  Target fluxes (total flux
into a given state) require no such normalization.

-----------------------------------------------------------------------------
Command-line options
-----------------------------------------------------------------------------
'''

    def __init__(self, parent):
        super().__init__(parent)

    def open_files(self):
        self.output_file = h5io.WESTPAH5File(self.output_filename, 'a', creating_program=True)
        h5io.stamp_creator_data(self.output_file)
        self.assignments_file = h5io.WESTPAH5File(self.assignments_filename, 'r')  # , driver='core', backing_store=False)
        if not self.iter_range.check_data_iter_range_least(self.assignments_file):
            raise ValueError('assignments data do not span the requested iterations')

    def go(self):
        pi = self.progress.indicator
        with pi:
            self.w_kinetics()


# The old w_kinavg
class DKinAvg(AverageCommands):
    subcommand = 'kinetics'
    help_text = 'Generates rate and flux values from a WESTPA simulation via tracing.'
    default_kinetics_file = 'direct.h5'
    description = '''\
Calculate average rates/fluxes and associated errors from weighted ensemble
data. Bin assignments (usually "assign.h5") and kinetics data (usually
"direct.h5") data files must have been previously generated (see
"w_assign --help" and "w_direct init --help" for information on
generating these files).

The evolution of all datasets may be calculated, with or without confidence
intervals.

-----------------------------------------------------------------------------
Output format
-----------------------------------------------------------------------------

The output file (-o/--output, usually "direct.h5") contains the following
dataset:

  /avg_rates [state,state]
    (Structured -- see below) State-to-state rates based on entire window of
    iterations selected.

  /avg_total_fluxes [state]
    (Structured -- see below) Total fluxes into each state based on entire
    window of iterations selected.

  /avg_conditional_fluxes [state,state]
    (Structured -- see below) State-to-state fluxes based on entire window of
    iterations selected.

If --evolution-mode is specified, then the following additional datasets are
available:

  /rate_evolution [window][state][state]
    (Structured -- see below). State-to-state rates based on windows of
    iterations of varying width.  If --evolution-mode=cumulative, then
    these windows all begin at the iteration specified with
    --start-iter and grow in length by --step-iter for each successive
    element. If --evolution-mode=blocked, then these windows are all of
    width --step-iter (excluding the last, which may be shorter), the first
    of which begins at iteration --start-iter.

  /target_flux_evolution [window,state]
    (Structured -- see below). Total flux into a given macro state based on
    windows of iterations of varying width, as in /rate_evolution.

  /conditional_flux_evolution [window,state,state]
    (Structured -- see below). State-to-state fluxes based on windows of
    varying width, as in /rate_evolution.

The structure of these datasets is as follows:

  iter_start
    (Integer) Iteration at which the averaging window begins (inclusive).

  iter_stop
    (Integer) Iteration at which the averaging window ends (exclusive).

  expected
    (Floating-point) Expected (mean) value of the observable as evaluated within
    this window, in units of inverse tau.

  ci_lbound
    (Floating-point) Lower bound of the confidence interval of the observable
    within this window, in units of inverse tau.

  ci_ubound
    (Floating-point) Upper bound of the confidence interval of the observable
    within this window, in units of inverse tau.

  stderr
    (Floating-point) The standard error of the mean of the observable
    within this window, in units of inverse tau.

  corr_len
    (Integer) Correlation length of the observable within this window, in units
    of tau.

Each of these datasets is also stamped with a number of attributes:

  mcbs_alpha
    (Floating-point) Alpha value of confidence intervals. (For example,
    *alpha=0.05* corresponds to a 95% confidence interval.)

  mcbs_nsets
    (Integer) Number of bootstrap data sets used in generating confidence
    intervals.

  mcbs_acalpha
    (Floating-point) Alpha value for determining correlation lengths.


-----------------------------------------------------------------------------
Command-line options
-----------------------------------------------------------------------------
'''

    def w_kinavg(self):
        pi = self.progress.indicator
        # pi = None

        # We're initializing the various datasets...
        self.open_files()
        self.open_assignments()
        # Obviously, this is for the conditional and total fluxes.  This is really all we need to sort for this.
        cond_fluxes = h5io.IterBlockedDataset(self.kinetics_file['conditional_fluxes'])
        cond_fluxes.cache_data()
        total_fluxes = h5io.IterBlockedDataset(self.kinetics_file['total_fluxes'])

        # This is necessary for both color and state populations...
        # ... but we also need this for the kinetics calculations.
        pops = h5io.IterBlockedDataset(self.assignments_file['labeled_populations'])
        pops.cache_data()
        pops.data = pops.data.sum(axis=2)

        submit_kwargs = dict(
            pi=pi, nstates=self.nstates, start_iter=self.start_iter, stop_iter=self.stop_iter, step_iter=self.step_iter
        )

        # Calculate averages for the simulation, then report, if necessary.

        submit_kwargs['dataset'] = {'dataset': cond_fluxes, 'pops': pops}
        avg_rates = self.run_calculation(
            eval_block=_rate_eval_block, name='Rate Evolution', dim=2, do_averages=True, **submit_kwargs
        )
        self.output_file.replace_dataset('avg_rates', data=avg_rates[1])

        submit_kwargs['dataset'] = {'dataset': cond_fluxes}
        avg_conditional_fluxes = self.run_calculation(
            eval_block=_2D_simple_eval_block, name='Conditional Flux Evolution', dim=2, do_averages=True, **submit_kwargs
        )
        self.output_file.replace_dataset('avg_conditional_fluxes', data=avg_conditional_fluxes[1])

        submit_kwargs['dataset'] = {'dataset': total_fluxes}
        avg_total_fluxes = self.run_calculation(
            eval_block=_1D_simple_eval_block, name='Target Flux Evolution', dim=1, do_averages=True, **submit_kwargs
        )
        self.output_file.replace_dataset('avg_total_fluxes', data=avg_total_fluxes[1])

        # Now, print them!

        # We've returned an average, but it still exists in a timeslice format.  So we need to return the 'last' value.
        if self.display_averages:
            self.print_averages(avg_total_fluxes[1], '\nfluxes into macrostates:', dim=1)
            self.print_averages(avg_conditional_fluxes[1], '\nfluxes from state to state:', dim=2)
            self.print_averages(avg_rates[1], '\nrates from state to state:', dim=2)

        # Do a bootstrap evolution.
        submit_kwargs['dataset'] = {'dataset': cond_fluxes, 'pops': pops}
        rate_evol = self.run_calculation(eval_block=_rate_eval_block, name='Rate Evolution', dim=2, **submit_kwargs)
        self.output_file.replace_dataset('rate_evolution', data=rate_evol, shuffle=True, compression=9)

        submit_kwargs['dataset'] = {'dataset': cond_fluxes}
        rate_evol = self.run_calculation(
            eval_block=_2D_simple_eval_block, name='Conditional Flux Evolution', dim=2, **submit_kwargs
        )
        self.output_file.replace_dataset('conditional_flux_evolution', data=rate_evol, shuffle=True, compression=9)

        submit_kwargs['dataset'] = {'dataset': total_fluxes}
        rate_evol = self.run_calculation(eval_block=_1D_simple_eval_block, name='Target Flux Evolution', dim=1, **submit_kwargs)
        self.output_file.replace_dataset('target_flux_evolution', data=rate_evol, shuffle=True, compression=9)

    def go(self):
        pi = self.progress.indicator
        with pi:
            self.w_kinavg()


# The old w_stateprobs
class DStateProbs(AverageCommands):
    subcommand = 'probs'
    help_text = 'Calculates color and state probabilities via tracing.'
    default_kinetics_file = 'direct.h5'
    description = '''\
Calculate average populations and associated errors in state populations from
weighted ensemble data. Bin assignments, including macrostate definitions,
are required. (See "w_assign --help" for more information).

-----------------------------------------------------------------------------
Output format
-----------------------------------------------------------------------------

The output file (-o/--output, usually "direct.h5") contains the following
dataset:

  /avg_state_probs [state]
    (Structured -- see below) Population of each state across entire
    range specified.

  /avg_color_probs [state]
    (Structured -- see below) Population of each ensemble across entire
    range specified.

If --evolution-mode is specified, then the following additional datasets are
available:

  /state_pop_evolution [window][state]
    (Structured -- see below). State populations based on windows of
    iterations of varying width.  If --evolution-mode=cumulative, then
    these windows all begin at the iteration specified with
    --start-iter and grow in length by --step-iter for each successive
    element. If --evolution-mode=blocked, then these windows are all of
    width --step-iter (excluding the last, which may be shorter), the first
    of which begins at iteration --start-iter.

  /color_prob_evolution [window][state]
    (Structured -- see below). Ensemble populations based on windows of
    iterations of varying width.  If --evolution-mode=cumulative, then
    these windows all begin at the iteration specified with
    --start-iter and grow in length by --step-iter for each successive
    element. If --evolution-mode=blocked, then these windows are all of
    width --step-iter (excluding the last, which may be shorter), the first
    of which begins at iteration --start-iter.

The structure of these datasets is as follows:

  iter_start
    (Integer) Iteration at which the averaging window begins (inclusive).

  iter_stop
    (Integer) Iteration at which the averaging window ends (exclusive).

  expected
    (Floating-point) Expected (mean) value of the observable as evaluated within
    this window, in units of inverse tau.

  ci_lbound
    (Floating-point) Lower bound of the confidence interval of the observable
    within this window, in units of inverse tau.

  ci_ubound
    (Floating-point) Upper bound of the confidence interval of the observable
    within this window, in units of inverse tau.

  stderr
    (Floating-point) The standard error of the mean of the observable
    within this window, in units of inverse tau.

  corr_len
    (Integer) Correlation length of the observable within this window, in units
    of tau.

Each of these datasets is also stamped with a number of attributes:

  mcbs_alpha
    (Floating-point) Alpha value of confidence intervals. (For example,
    *alpha=0.05* corresponds to a 95% confidence interval.)

  mcbs_nsets
    (Integer) Number of bootstrap data sets used in generating confidence
    intervals.

  mcbs_acalpha
    (Floating-point) Alpha value for determining correlation lengths.


-----------------------------------------------------------------------------
Command-line options
-----------------------------------------------------------------------------
'''

    def calculate_state_populations(self, pops):
        # ... but then this is how the state populations are done.
        # This was taken, more or less, from the old w_stateprobs
        iter_count = self.stop_iter - self.start_iter
        all_state_pops = np.empty((iter_count, self.nstates + 1), weight_dtype)
        iter_state_pops = np.empty((self.nstates + 1,), weight_dtype)
        avg_state_pops = np.zeros((self.nstates + 1,), weight_dtype)
        pops.cache_data(max_size='available')
        state_map = self.assignments_file['state_map'][...]
        try:
            for iiter, n_iter in enumerate(range(self.start_iter, self.stop_iter)):
                iter_state_pops.fill(0)
                labeled_pops = pops.iter_entry(n_iter)
                accumulate_state_populations_from_labeled(labeled_pops, state_map, iter_state_pops, check_state_map=False)
                all_state_pops[iiter] = iter_state_pops
                avg_state_pops += iter_state_pops
                del labeled_pops
        finally:
            pops.drop_cache()

        state_pops = h5io.IterBlockedDataset.empty_like(pops)
        state_pops.data = all_state_pops
        return state_pops

    def w_stateprobs(self):
        pi = self.progress.indicator

        self.open_files()
        self.open_assignments()
        # So far, we definitely need this boilerplate...
        # pi.new_operation('Reading data')

        # This is necessary for both color and state populations...
        pops = h5io.IterBlockedDataset(self.assignments_file['labeled_populations'])

        state_pops = self.calculate_state_populations(pops)

        # This now sorts it for the color populations
        pops.cache_data()
        pops.data = pops.data.sum(axis=2)

        submit_kwargs = dict(
            pi=pi,
            nstates=self.nstates,
            start_iter=self.start_iter,
            stop_iter=self.stop_iter,
            step_iter=self.step_iter,
            eval_block=_1D_simple_eval_block,
        )

        # Calculate and print averages
        submit_kwargs['dataset'] = {'dataset': pops}
        color_evol_avg = self.run_calculation(name='Color Probability Evolution', dim=1, do_averages=True, **submit_kwargs)
        self.output_file.replace_dataset('avg_color_probs', data=color_evol_avg[1], shuffle=True, compression=9)

        submit_kwargs['dataset'] = {'dataset': state_pops}
        state_evol_avg = self.run_calculation(name='State Probability Evolution', dim=1, do_averages=True, **submit_kwargs)
        self.output_file.replace_dataset(name='avg_state_probs', data=state_evol_avg[1], shuffle=True, compression=9)

        # Print!
        if self.display_averages:
            self.print_averages(color_evol_avg[1], '\naverage color probabilities:', dim=1)
            self.print_averages(state_evol_avg[1], '\naverage state probabilities:', dim=1)

        # Now, do a bootstrap evolution
        submit_kwargs['dataset'] = {'dataset': pops}
        pop_evol = self.run_calculation(name='Color Probability Evolution', dim=1, **submit_kwargs)
        self.output_file.replace_dataset('color_prob_evolution', data=pop_evol, shuffle=True, compression=9)

        submit_kwargs['dataset'] = {'dataset': state_pops}
        pop_evol = self.run_calculation(name='State Probability Evolution', dim=1, **submit_kwargs)
        self.output_file.replace_dataset(name='state_pop_evolution', data=pop_evol, shuffle=True, compression=9)

    def go(self):
        pi = self.progress.indicator
        with pi:
            self.w_stateprobs()


# Just a convenience class to run everything.
class DAll(DStateProbs, DKinAvg, DKinetics):
    subcommand = 'all'
    help_text = 'Runs the full suite, including the tracing of events.'
    default_kinetics_file = 'direct.h5'
    description = '''\
A convenience function to run init/kinetics/probs. Bin assignments,
including macrostate definitions, are required. (See
"w_assign --help" for more information).

For more information on the individual subcommands this subs in for, run
w_direct {init/kinetics/probs} --help.

-----------------------------------------------------------------------------
Command-line options
-----------------------------------------------------------------------------
'''

    def go(self):
        # One minor issue; as this stands now, since it's inheriting from all the other classes, it needs
        # a kinetics file to instantiate the other attributes.  We'll need to modify how the loading works, there.
        pi = self.progress.indicator
        with pi:
            self.w_kinetics()
            self.w_kinavg()
            self.w_stateprobs()


# Just a convenience class to average the observables.
class DAverage(DStateProbs, DKinAvg):
    subcommand = 'average'
    help_text = 'Averages and returns fluxes, rates, and color/state populations.'
    default_kinetics_file = 'direct.h5'
    description = '''\
A convenience function to run kinetics/probs. Bin assignments,
including macrostate definitions, are required. (See
"w_assign --help" for more information).

For more information on the individual subcommands this subs in for, run
w_direct {kinetics/probs} --help.

-----------------------------------------------------------------------------
Command-line options
-----------------------------------------------------------------------------
'''

    def go(self):
        pi = self.progress.indicator
        with pi:
            self.w_kinavg()
            self.w_stateprobs()


class WDirect(WESTMasterCommand, WESTParallelTool):
    prog = 'w_direct'
    # subcommands = [AvgTraceSubcommand,AvgMatrixSubcommand]
    subcommands = [DKinetics, DAverage, DKinAvg, DStateProbs, DAll]
    subparsers_title = 'direct kinetics analysis schemes'


def entry_point():
    WDirect().main()


if __name__ == '__main__':
    entry_point()



import h5py
import logging
import operator
import numpy as np

import westpa
from westpa.cli.core import w_init
from westpa.cli.core import w_run
from westpa.core.extloader import get_object
from westpa.core.segment import Segment
from westpa import analysis

import json

import os
import shutil
import pickle
import importlib.util

import tqdm

import mdtraj as md
from rich.logging import RichHandler

from matplotlib import pyplot as plt

# Ensure this is installed via pip. msm_we's setup.py is all set up for that.
# Navigate to the folder where msm_we is, and run python3 -m pip install .
#   If you're doing development on msm_we, add the -e flag to pip, i.e. "python3 -m pip install -e ."
#   -e will install it in editable mode, so changes to msm_we will take effect next time it's imported.
#   Otherwise, if you modify the msm_we code, you'll need to re-install it through pip.
from msm_we import msm_we

import ray
import tempfile

EPS = np.finfo(np.float64).eps

log = logging.getLogger(__name__)
log.setLevel("INFO")
log.propagate = False
log.addHandler(RichHandler())

msm_we_logger = logging.getLogger("msm_we.msm_we")
msm_we_logger.setLevel("INFO")

# Map structure types to extensions.
# This tells the plugin what extension to put on generated start-state files.
STRUCT_EXTENSIONS = {
    md.formats.PDBTrajectoryFile: "pdb",
    md.formats.AmberRestartFile: "rst7",
}

EXTENSION_LOCKFILE = 'doing_extension'


def check_target_reached(h5_filename):
    """
    Check if the target state was reached, given the data in a WEST H5 file.

    Parameters
    ----------
    h5_filename: string
        Path to a WESTPA HDF5 data file
    """
    with h5py.File(h5_filename, 'r') as h5_file:
        # Get the key to the final iteration. Need to do -2 instead of -1 because there's an empty-ish final iteration
        #   written.
        for iteration_key in list(h5_file['iterations'].keys())[-2:0:-1]:
            endpoint_types = h5_file[f'iterations/{iteration_key}/seg_index']['endpoint_type']
            if Segment.SEG_ENDPOINT_RECYCLED in endpoint_types:
                log.debug(f"recycled segment found in file {h5_filename} at iteration {iteration_key}")
                return True
    return False


def fix_deprecated_initialization(initialization_state):
    """
    I changed my initialization JSON schema to use underscores instead of hyphens so I can directly expand it into
    keywords arguments to w_init. This just handles any old-style JSON files I still had, so they don't choke and die.
    """

    log.debug(f"Starting processing, dict is now {initialization_state}")

    # Some of my initial files had this old-style formatting. Handle it for now, but remove eventually
    for old_key, new_key in [
        ('tstate-file', 'tstate_file'),
        ('bstate-file', 'bstate_file'),
        ('sstate-file', 'sstate_file'),
        ('segs-per-state', 'segs_per_state'),
    ]:
        if old_key in initialization_state.keys():
            log.warning(
                f"This initialization JSON file uses the deprecated  " f"hyphenated form for  {old_key}. Replace with underscores."
            )

            value = initialization_state.pop(old_key)
            initialization_state[new_key] = value

    log.debug(f"Finished processing, dict is now {initialization_state}")
    return initialization_state


# TODO: Break this out into a separate module, let it be specified (if it's necessary) as a plugin option
#   This may not always be required -- i.e. you may be able to directly output to the h5 file in your propagator
def prepare_coordinates(plugin_config, h5file, we_h5filename):
    """
    Copy relevant coordinates from trajectory files into <iteration>/auxdata/coord of the h5 file.

    Directly modifies the input h5 file.

    Adds ALL coordinates to auxdata/coord.

    Adapted from original msmWE collectCoordinates.py script.

    Parameters
    ----------
    plugin_config: YAMLConfig object
        Stores the configuration options provided to the plugin in the WESTPA configuration file

    h5file: h5py.File
        WESTPA h5 data file

    we_h5filename: string
        Name of the WESTPA h5 file
    """

    refPDBfile = plugin_config.get('ref_pdb_file')
    modelName = plugin_config.get('model_name')

    # TODO: Don't need this explicit option, use WEST_SIM_ROOT or something
    WEfolder = plugin_config.get('we_folder')

    parentTraj = plugin_config.get('parent_traj_filename')
    childTraj = plugin_config.get('child_traj_filename')
    pcoord_ndim = plugin_config.get('pcoord_ndim', 1)

    model = msm_we.modelWE()
    log.info('Preparing coordinates...')

    # Only need the model to get the number of iterations and atoms
    # TODO: Replace this with something more lightweight, get directly from WE
    log.debug(f'Doing collectCoordinates on  WE file {we_h5filename}')
    model.initialize(
        [we_h5filename],
        refPDBfile,
        modelName,
        # Pass some dummy arguments -- these aren't important, this model is just created for convenience
        # in the coordinate collection. Dummy arguments prevent warnings from being raised.
        basis_pcoord_bounds=None,
        target_pcoord_bounds=None,
        tau=1,
        pcoord_ndim=pcoord_ndim,
        _suppress_boundary_warning=True,
    )
    model.get_iterations()

    log.debug(f"Found {model.maxIter} iterations")

    n_iter = None
    for n_iter in tqdm.tqdm(range(1, model.maxIter + 1)):

        nS = model.numSegments[n_iter - 1].astype(int)
        coords = np.zeros((nS, 2, model.nAtoms, 3))
        dsetName = "/iterations/iter_%08d/auxdata/coord" % int(n_iter)

        coords_exist = False
        try:
            dset = h5file.create_dataset(dsetName, np.shape(coords))
        except (RuntimeError, ValueError):
            log.debug('coords exist for iteration ' + str(n_iter) + ' NOT overwritten')
            coords_exist = True
            continue

        for iS in range(nS):
            trajpath = WEfolder + "/traj_segs/%06d/%06d" % (n_iter, iS)

            try:
                coord0 = np.squeeze(md.load(f'{trajpath}/{parentTraj}', top=model.reference_structure.topology)._xyz)
            except OSError:
                log.warning("Parent traj file doesn't exist, loading reference structure coords")
                coord0 = np.squeeze(model.reference_structure._xyz)

            coord1 = np.squeeze(md.load(f'{trajpath}/{childTraj}', top=model.reference_structure.topology)._xyz)

            coords[iS, 0, :, :] = coord0
            coords[iS, 1, :, :] = coord1

        if not coords_exist:
            dset[:] = coords

    log.debug(f"Wrote coords for {n_iter} iterations.")


def msmwe_compute_ss(plugin_config, west_files):
    """
    Prepare and initialize an msm_we model, and use it to predict a steady-state distribution.

    1. Load coordinate data
    2. Perform dimensionality reduction
    3. Compute flux and transition matrices
    4. Compute steady-state distribution (via eigenvectors of transition matrix)
    5. Compute target-state flux

    TODO
    ----
    This function does far too many things. Break it up a bit.

    Parameters
    ----------
    plugin_config: YAMLConfig object
        Stores the configuration options provided to the plugin in the WESTPA configuration file

    last_iter: int
        The last WE iteration to use for computing steady-state.

    Returns
    -------
    ss_alg: np.ndarray
        The steady-state distribution

    ss_flux: float
        Flux into target state

    model: modelWE object
        The modelWE object produced for analysis.
    """

    n_lag = 0

    log.debug("Initializing msm_we")

    # TODO: Refactor this to use westpa.core.extloader.get_object
    #   I'm reinventing the wheel a bit here, I can replace almost all this code w/ that
    # ##### Monkey-patch modelWE with the user-override functions
    override_file = plugin_config.get('user_functions')

    # First, import the file with the user-override functions
    # This is a decently janky implementation, but it seems to work, and I don't know of a better way of doing it.
    # This is nice because it avoids mucking around with the path, which I think is a Good Thing.

    # We're given a path to the user-specified file containing overrides
    # This comes from https://stackoverflow.com/questions/67631/how-to-import-a-module-given-the-full-path

    # I don't think the name provided here actually matters
    user_override_spec = importlib.util.spec_from_file_location("override_module", override_file)
    user_overrides = importlib.util.module_from_spec(user_override_spec)

    # Make the functions that were overriden in override_file available in the namespace under user_overrides
    user_override_spec.loader.exec_module(user_overrides)

    # So now we can do the actual monkey-patching of modelWE.
    # We monkey-patch at the module level rather than just override the function in the instanced object
    #   so that the functions retain access to self.
    msm_we.modelWE.processCoordinates = user_overrides.processCoordinates
    # ##### Done with monkey-patching.

    model = msm_we.modelWE()
    streaming = plugin_config.get('streaming', False)

    refPDBfile = plugin_config.get('ref_pdb_file')
    modelName = plugin_config.get('model_name')
    n_clusters = plugin_config.get('n_clusters')
    tau = plugin_config.get('tau', None)
    pcoord_ndim = plugin_config.get('pcoord_ndim', 1)

    basis_pcoord_bounds = np.array(plugin_config.get('basis_pcoord_bounds', np.nan), dtype=float)
    target_pcoord_bounds = np.array(plugin_config.get('target_pcoord_bounds', np.nan), dtype=float)

    if np.isnan(basis_pcoord_bounds).any() or np.isnan(target_pcoord_bounds).any():
        log.critical(
            "Target and/or basis pcoord bounds were not specified. "
            "Set them using the 'basis_pcoord_bounds' or 'target_pcoord_bounds' parameters. "
            "'basis/target_pcoord1_min/max' and 'basis/target_pcoord1' are no longer supported. "
            "See https://jdrusso.github.io/msm_we/api.html#msm_we.msm_we.modelWE.initialize for details."
        )

    if tau is None:
        log.warning('No tau provided to restarting plugin. Defaulting to 1.')
        tau = 1

    # Fire up the model object
    model.initialize(
        west_files,
        refPDBfile,
        modelName,
        basis_pcoord_bounds=basis_pcoord_bounds,
        target_pcoord_bounds=target_pcoord_bounds,
        tau=tau,
        pcoord_ndim=pcoord_ndim,
    )

    model.dimReduceMethod = plugin_config.get('dim_reduce_method')

    model.n_lag = n_lag

    log.debug("Loading in iteration data.. (this could take a while)")

    # First dimension is the total number of segments
    model.get_iterations()

    model.get_coordSet(model.maxIter)

    model.dimReduce()

    first_iter, last_iter = model.first_iter, model.maxIter

    clusterFile = modelName + "_clusters_s" + str(first_iter) + "_e" + str(last_iter) + "_nC" + str(n_clusters) + ".h5"
    # TODO: Uncomment this to actually load the clusterFile if it exists.  For now, disable for development.
    exists = os.path.isfile(clusterFile)
    exists = False
    log.warning("Skipping any potential cluster reloading!")

    log.info(f"Launching Ray with {plugin_config.get('n_cpus', 1)} cpus")

    ray_tempdir_root = plugin_config.get('ray_temp_dir', None)
    if ray_tempdir_root is not None:
        ray_tempdir = tempfile.TemporaryDirectory(dir=ray_tempdir_root)
        log.info(f"Using {ray_tempdir.name} as temp_dir for Ray")
        ray.init(
            num_cpus=plugin_config.get('n_cpus', 1), _temp_dir=ray_tempdir.name, ignore_reinit_error=True, include_dashboard=False
        )
    else:
        ray.init(num_cpus=plugin_config.get('n_cpus', 1), ignore_reinit_error=True, include_dashboard=False)

    # If a cluster file with the name corresponding to these parameters exists, load clusters from it.
    if exists:
        log.debug("loading clusters...")
        model.load_clusters(clusterFile)
    # Otherwise, do the clustering (which will create and save to that file)
    else:
        # FIXME: This gives the wrong shape, but loading from the clusterfile gives the right shape
        log.debug("clustering coordinates into " + str(n_clusters) + " clusters...")
        model.cluster_coordinates(n_clusters, streaming=streaming)

    first_iter = 1
    model.get_fluxMatrix(n_lag, first_iter, last_iter)  # extracts flux matrix, output model.fluxMatrixRaw
    log.debug(f"Unprocessed flux matrix has shape {model.fluxMatrixRaw.shape}")
    model.organize_fluxMatrix()  # gets rid of bins with no connectivity, sorts along p1, output model.fluxMatrix
    model.get_Tmatrix()  # normalizes fluxMatrix to transition matrix, output model.Tmatrix

    log.debug(f"Processed flux matrix has shape {model.fluxMatrix.shape}")

    model.get_steady_state()  # gets steady-state from eigen decomp, output model.pSS
    model.get_steady_state_target_flux()  # gets steady-state target flux, output model.JtargetSS

    # Why is model.pss sometimes the wrong shape? It's "occasionally" returned as a nested array.
    # Squeeze fixes it and removes the dimension of length 1, but why does it happen in the first place?

    if type(model.pSS) is np.matrix:
        ss_alg = np.squeeze(model.pSS.A)
    else:
        ss_alg = np.squeeze(model.pSS)

    ss_flux = model.JtargetSS

    log.debug("Got steady state:")
    log.debug(ss_alg)
    log.debug(ss_flux)

    log.info("Completed flux matrix calculation and steady-state estimation!")

    log.info("Starting block validation")
    num_validation_groups = plugin_config.get('n_validation_groups', 2)
    num_validation_blocks = plugin_config.get('n_validation_blocks', 4)
    try:
        model.do_block_validation(num_validation_groups, num_validation_blocks, use_ray=True)
    except Exception as e:
        log.exception(e)
        log.error("Failed block validation! Continuing with restart, but BEWARE!")
    ray.shutdown()

    return ss_alg, ss_flux, model


class RestartDriver:
    """
    WESTPA plugin to automatically handle estimating steady-state from a WE run, re-initializing a new WE run in that
    steady-state, and then running that initialized WE run.

    Data from the previous run will be stored in the restart<restart_number>/ subdirectory of $WEST_SIM_ROOT.

    This plugin depends on having the start-states implementation in the main WESTPA code, which allows initializing
    a WE run using states that are NOT later used for recycling.

    These are used so that when the new WE run is initialized, initial structure selection is chosen by w_init, using
    weights assigned to the start-states based on MSM bin weight and WE segment weight.

    Since it closes out the current WE run and starts a new one, this plugin should run LAST, after all other plugins.
    """

    def __init__(self, sim_manager, plugin_config):
        """
        Initialize the RestartDriver plugin.

        Pulls the data_manager and sim_manager from the WESTPA run that just completed, along with
        """

        westpa.rc.pstatus("Restart plugin initialized")

        if not sim_manager.work_manager.is_master:
            westpa.rc.pstatus("Reweighting not master, skipping")
            return

        self.data_manager = sim_manager.data_manager
        self.sim_manager = sim_manager

        self.plugin_config = plugin_config

        self.restart_file = plugin_config.get('restart_file', 'restart.dat')
        self.initialization_file = plugin_config.get('initialization_file', 'restart_initialization.json')

        self.extension_iters = plugin_config.get('extension_iters', 0)
        self.max_total_iterations = westpa.rc.config.get(['west', 'propagation', 'max_total_iterations'], default=None)
        self.base_total_iterations = self.max_total_iterations

        self.coord_len = plugin_config.get('coord_len', 2)
        self.n_restarts = plugin_config.get('n_restarts', -1)
        self.n_runs = plugin_config.get('n_runs', 1)

        # Number of CPUs available for parallelizing msm_we calculations
        self.parallel_cpus = plugin_config.get('n_cpus', 1)
        self.ray_tempdir = plugin_config.get('ray_temp_dir', None)

        # .get() might return this as a bool anyways, but be safe
        self.debug = bool(plugin_config.get('debug', False))
        if self.debug:
            log.setLevel("DEBUG")
            msm_we_logger.setLevel("DEBUG")

        # Default to using all restarts
        self.restarts_to_use = plugin_config.get('n_restarts_to_use', self.n_restarts)
        assert self.restarts_to_use > 0 or self.restarts_to_use == -1, "Invalid number of restarts to use"
        if self.restarts_to_use >= 1:
            assert (
                self.restarts_to_use == self.restarts_to_use // 1
            ), "If choosing a decimal restarts_to_use, must be between 0 and 1."

        struct_filetype = plugin_config.get('struct_filetype', 'mdtraj.formats.PDBTrajectoryFile')
        self.struct_filetype = get_object(struct_filetype)

        # This should be low priority, because it closes the H5 file and starts a new WE run. So it should run LAST
        #   after any other plugins.
        self.priority = plugin_config.get('priority', 100)  # I think a big number is lower priority...

        sim_manager.register_callback(sim_manager.finalize_run, self.prepare_new_we, self.priority)

        # Initialize data
        self.ss_alg = None
        self.ss_dist = None
        self.model = None

    def get_original_bins(self):
        """
        Obtains the WE bins and their probabilities at the end of the previous iteration.

        Returns
        -------
        bins : np.ndarray
            Array of WE bins

        binprobs: np.ndarray
            WE bin weights
        """

        we_driver = self.sim_manager.we_driver
        bins = we_driver.next_iter_binning
        n_bins = len(bins)
        binprobs = np.fromiter(map(operator.attrgetter('weight'), bins), dtype=np.float64, count=n_bins)

        return bins, binprobs

    @property
    def cur_iter(self):
        """
        Get the current WE iteration.

        Returns
        -------
        int: The current iteration. Subtract one, because in finalize_run the iter has been incremented
        """
        return self.sim_manager.n_iter - 1

    @property
    def is_last_iteration(self):
        """
        Get whether this is, or is past, the last iteration in this WE run.

        Returns
        -------
        bool: Whether the current iteration is the final iteration
        """

        final_iter = self.sim_manager.max_total_iterations

        return self.cur_iter >= final_iter

    def prepare_extension_run(self, run_number, restart_state, first_extension=False):
        """
        Copy the necessary files for an extension run  (versus initializing a fresh run)

        Parameters
        ----------
        run_number: int
            The index of this run (should be 1-indexed!)

        restart_state: dict
            Dictionary holding the current state of the restarting procedure

        first_extension: bool
            True if this is the first run of an extension set. If True, then back up west.cfg, and write the extended
            west.cfg file.
        """

        log.debug(f"Linking run files from restart0/run{run_number}")

        # Copy traj_segs, seg_logs, and west.h5 for restart0/runXX back into ./
        #       Later: (May only need to copy the latest iteration traj_segs, to avoid tons of back and forth)
        try:
            shutil.rmtree('traj_segs')
            shutil.rmtree('seg_logs')
        except OSError as e:
            if str(e) == 'Cannot call rmtree on a symbolic link':
                os.unlink('traj_segs')
                os.unlink('seg_logs')

        os.remove(self.data_manager.we_h5filename)

        os.symlink(f'restart0/run{run_number}/traj_segs', 'traj_segs')
        os.symlink(f'restart0/run{run_number}/seg_logs', 'seg_logs')

        if first_extension:

            # Get lines to make a new west.cfg by extending west.propagation.max_total_iterations
            with open('west.cfg', 'r') as west_config:
                lines = west_config.readlines()
                for i, line in enumerate(lines):
                    # Parse out the number of maximum iterations
                    if 'max_total_iterations' in line:
                        max_iters = [int(i) for i in line.replace(':', ' ').replace('\n', ' ').split() if i.isdigit()]
                        new_max_iters = max_iters[0] + self.extension_iters
                        new_line = f"{line.split(':')[0]}: {new_max_iters}\n"
                        lines[i] = new_line
                        break

        with open(self.restart_file, 'w') as fp:
            json.dump(restart_state, fp)

        log.info("First WE extension run ready!")
        westpa.rc.pstatus(
            f"\n\n===== Restart {restart_state['restarts_completed']}, "
            + f"Run {restart_state['runs_completed'] + 1} extension running =====\n"
        )

        # TODO: I can't just go straight into a w_run here. w_run expects some things to be set I think, that aren't.
        #   I can do w_init, and do a new simulation just fine...
        #   I can do this on ONE run repeatedly just fine
        #   But if I try to just copy files and continue like this, there's something screwy in state somewhere that
        #       causes it to fail.
        #   The error has to do with offsets in the HDF5 file?
        #   Need to figure out what state would be cleared by w_init

        # Frankly, this is a really sketchy way of doing this, but it seems to work...
        #   I remain skeptical there's not something weird under the hood that isn't being addressed correctly with
        #   regard to state, but if it works, it's good enough for now..
        westpa.rc.sim_manager.segments = None
        shutil.copy(f'restart0/run{run_number}/west.h5', self.data_manager.we_h5filename)
        self.data_manager.open_backing()

        log.debug(f"Sim manager thought n_iter was {westpa.rc.sim_manager.n_iter}")
        log.debug(f"Data manager thought current_iteration was {self.data_manager.current_iteration}")
        log.debug(f"{self.sim_manager} vs {westpa.rc.sim_manager}")

        if run_number == 1:
            westpa.rc.sim_manager.max_total_iterations += self.extension_iters

        w_run.run_simulation()
        return

    def generate_plots(self, restart_directory):

        model = self.model

        log.info("Producing flux-profile, pseudocommittor, and target flux comparison plots.")
        flux_pcoord_fig, flux_pcoord_ax = plt.subplots()
        model.plot_flux(ax=flux_pcoord_ax, suppress_validation=True)
        flux_pcoord_fig.text(x=0.1, y=-0.05, s='This flux profile should become flatter after restarting', fontsize=12)
        flux_pcoord_ax.legend(bbox_to_anchor=(1.01, 1.0), loc="upper left")
        flux_pcoord_fig.savefig(f'{restart_directory}/flux_plot.pdf', bbox_inches="tight")

        flux_pseudocomm_fig, flux_pseudocomm_ax = plt.subplots()
        model.plot_flux_committor(ax=flux_pseudocomm_ax, suppress_validation=True)
        flux_pseudocomm_fig.text(
            x=0.1,
            y=-0.05,
            s='This flux profile should become flatter after restarting.'
            '\nThe x-axis is a "pseudo"committor, since it may be '
            'calculated from WE trajectories in the one-way ensemble.',
            fontsize=12,
        )
        flux_pseudocomm_ax.legend(bbox_to_anchor=(1.01, 1.0), loc="upper left")
        flux_pseudocomm_fig.savefig(f'{restart_directory}/pseudocomm-flux_plot.pdf', bbox_inches="tight")

        flux_comparison_fig, flux_comparison_ax = plt.subplots(figsize=(7, 3))
        # Get haMSM flux estimates
        models = [model]
        models.extend(model.validation_models)
        n_validation_models = len(model.validation_models)

        flux_estimates = []
        for _model in models:
            flux_estimates.append(_model.JtargetSS)

        hamsm_flux_colors = iter(plt.rcParams['axes.prop_cycle'].by_key()['color'])
        direct_flux_colors = iter(plt.cm.cool(np.linspace(0.2, 0.8, len(model.fileList))))

        # Get WE direct flux estimate
        for _file in model.fileList:

            run = analysis.Run(_file)
            last_iter = run.num_iterations
            recycled = list(run.iteration(last_iter - 1).recycled_walkers)
            target_flux = sum(walker.weight for walker in recycled) / model.tau

            # TODO: Correct for time!
            if len(_file) >= 15:
                short_filename = f"....{_file[-12:]}"
            else:
                short_filename = _file

            if target_flux == 0:
                continue

            flux_comparison_ax.axhline(
                target_flux,
                color=next(direct_flux_colors),
                label=f"Last iter WE direct {target_flux:.2e}" f"\n  ({short_filename})",
                linestyle='--',
            )

        flux_comparison_ax.axhline(
            flux_estimates[0], label=f"Main model estimate\n  {flux_estimates[0]:.2e}", color=next(hamsm_flux_colors)
        )
        for i in range(1, n_validation_models + 1):
            flux_comparison_ax.axhline(
                flux_estimates[i],
                label=f"Validation model {i - 1} estimate\n  {flux_estimates[i]:.2e}",
                color=next(hamsm_flux_colors),
            )

        flux_comparison_ax.legend(bbox_to_anchor=(1.01, 0.9), loc='upper left')
        flux_comparison_ax.set_yscale('log')
        flux_comparison_ax.set_ylabel('Flux')
        flux_comparison_ax.set_xticks([])
        flux_comparison_fig.tight_layout()
        flux_comparison_fig.savefig(f'{restart_directory}/hamsm_vs_direct_flux_comparison_plot.pdf', bbox_inches="tight")

    def prepare_new_we(self):
        """
        This function prepares a new WESTPA simulation using haMSM analysis to accelerate convergence.

        The marathon functionality does re-implement some of the functionality of w_multi_west.
        However, w_multi_west merges independent WE simulations, which may or may not be desirable.
        I think for the purposes of this, it's good to keep the runs completely independent until haMSM model building.
        Either that, or I'm just justifying not having known about w_multi_west when I wrote this. TBD.

        # TODO: Replace all manual path-building with pathlib

        The algorithm is as follows:
            1. Check to see if we've just completed the final iteration
            2. Handle launching multiple runs, if desired
            2. Build haMSM
            3. Obtain structures for each haMSM bin
            4. Make each structure a start-state, with probability set by (MSM-bin SS prob / # structures in bin)
            5. Potentially some renormalization?
            6. Start new WE simulation
        """

        # Do nothing if it's not the final iteration
        if not self.is_last_iteration:
            print(self.cur_iter)
            return

        log.debug("Final iteration, preparing restart")

        restart_state = {'restarts_completed': 0, 'runs_completed': 0}

        # Check for the existence of the extension lockfile here
        doing_extension = os.path.exists(EXTENSION_LOCKFILE)

        # Look for a restart.dat file to get the current state (how many restarts have been performed already)
        if os.path.exists(self.restart_file):
            with open(self.restart_file, 'r') as fp:
                restart_state = json.load(fp)

        # This is the final iteration of a run, so mark this run as completed
        restart_state['runs_completed'] += 1

        # Make the folder to store data for this marathon
        restart_directory = f"restart{restart_state['restarts_completed']}"
        run_directory = f"{restart_directory}/run{restart_state['runs_completed']}"
        if not os.path.exists(run_directory):
            os.makedirs(run_directory)

        # Write coordinates to h5
        prepare_coordinates(self.plugin_config, self.data_manager.we_h5file, self.data_manager.we_h5filename)

        for data_folder in ['traj_segs', 'seg_logs']:
            old_path = data_folder

            # If you're doing an extension, this will be a symlink. So no need to copy, just unlink it and move on
            if doing_extension and os.path.islink(old_path):
                log.debug('Unlinking symlink')
                os.unlink(old_path)
                os.mkdir(old_path)
                continue

            new_path = f"{run_directory}/{old_path}"

            log.debug(f"Moving {old_path} to {new_path}")

            if os.path.exists(new_path):
                log.info(f"{new_path} already exists. Removing and overwriting.")
                shutil.rmtree(new_path)

            try:
                os.rename(old_path, new_path)
            except FileNotFoundError:
                log.warning(f"Folder {old_path} was not found." "This may be normal, but check your configuration.")
            else:
                # Make a new data folder for the next run
                os.mkdir(old_path)

        last_run = restart_state['runs_completed'] >= self.n_runs
        last_restart = restart_state['restarts_completed'] >= self.n_restarts

        # We've just finished a run. Let's check if we have to do any more runs in this marathon before doing a restart.
        #   In the case of n_runs == 1, then we're just doing a single run and restarting it every so often.
        #   Otherwise, a marathon consists of multiple runs,  and restarts are performed between marathons.
        if last_run:
            log.info(f"All {self.n_runs} runs in this marathon completed.")

            if last_restart:
                log.info("All restarts completed! Performing final analysis.")

            else:
                log.info("Proceeding to prepare a restart.")

                # Duplicating this is  gross, but given the structure here, my options are either put it above these ifs
                #   entirely, meaning it'll be unnecessarily run at the end of the final restart, or duplicate it below.
                log.info("Preparing coordinates for this run.")

                # Now, continue on to haMSM calculation below.

        # If we have more runs left to do in this marathon, prepare them
        elif not last_run:

            log.info(f"Run {restart_state['runs_completed']}/{self.n_runs} completed.")

            # TODO: Initialize a new run, from the same configuration as this run was
            #   On the 1st run, I can write bstates/tstates/sstates into restart files, and use those for spawning
            #   subsequent runs in the marathon. That way, I don't make unnecessary copies of all those.
            # Basis and target states are unchanged. Can I get the original parameters passed to w_init?
            # Ideally, I should be able to call w_init with the exact same parameters that went to it the first time
            initialization_state = {
                'tstate_file': None,
                'bstate_file': None,
                'sstate_file': None,
                'tstates': None,
                'bstates': None,
                'sstates': None,
                'segs_per_state': None,
            }

            # TODO: Implement this, and get rid of the initialization_file usage right below. Placeholder for now.
            if restart_state['runs_completed'] == 1:

                # Get and write basis, target, start states and segs per state for this marathon to disk
                pass

            # Save the WESTPA h5 data from this run
            self.data_manager.finalize_run()
            shutil.copyfile('west.h5', f"{run_directory}/west.h5")

            # If this is a regular, fresh run (not an extension)
            if not doing_extension:
                if os.path.exists(self.initialization_file):
                    with open(self.initialization_file, 'r') as fp:
                        initialization_dict = json.load(fp)
                        initialization_dict = fix_deprecated_initialization(initialization_dict)
                        initialization_state.update(initialization_dict)
                else:
                    raise Exception(
                        "No initialization JSON file provided -- " "I don't know how to start new runs in this marathon."
                    )

                westpa.rc.pstatus(
                    f"\n\n===== Restart {restart_state['restarts_completed']}, "
                    + f"Run {restart_state['runs_completed']+1} initializing =====\n"
                )

                westpa.rc.pstatus(
                    f"\nRun: \n\t w_init --tstate-file {initialization_state['tstate_file']} "
                    + f"--bstate-file {initialization_state['bstate_file']} "
                    f"--sstate-file {initialization_state['sstate_file']} "
                    f"--segs-per-state {initialization_state['segs_per_state']}\n"
                )

                w_init.initialize(
                    **initialization_state,
                    shotgun=False,
                )

                with open(self.restart_file, 'w') as fp:
                    json.dump(restart_state, fp)

                log.info("New WE run ready!")
                westpa.rc.pstatus(
                    f"\n\n===== Restart {restart_state['restarts_completed']}, "
                    + f"Run {restart_state['runs_completed']+1} running =====\n"
                )

                w_run.run_simulation()
                return

            # If we're doing an extension set
            #   Instead of w_initting a new iteration, copy the files from restart0/runXX back into ./
            elif doing_extension:

                self.prepare_extension_run(run_number=restart_state['runs_completed'] + 1, restart_state=restart_state)
                return

        log.debug(f"{restart_state['restarts_completed']}/{self.n_restarts} restarts completed")

        # Build the haMSM
        log.debug("Initializing haMSM")

        # Need to write the h5 file and close it out, but I need to get the current bstates first.
        original_bstates = self.sim_manager.current_iter_bstates
        if original_bstates is None:
            original_bstates = self.data_manager.get_basis_states(self.sim_manager.n_iter - 1)

        assert original_bstates is not None, "Bstates are none in the current iteration"

        original_tstates = self.data_manager.get_target_states(self.cur_iter)

        # Flush h5 file writes and copy it to the run directory
        self.data_manager.finalize_run()
        shutil.copyfile(self.data_manager.we_h5filename, f"{run_directory}/west.h5")

        # Use all files in all restarts
        # Restarts index at 0, because there's  a 0th restart before you've... restarted anything.
        # Runs index at 1, because Run 1 is the first run.
        # TODO: Let the user pick last half or something in the plugin config.
        marathon_west_files = []
        # When doing the first restart, restarts_completed is 0 (because the first restart isn't complete yet) and
        #   the data generated during this restart is in /restart0.
        # So when doing the Nth restart, restarts_completed is N-1

        # If set to -1, use all restarts
        if self.restarts_to_use == -1:
            last_N_restarts = 1 + restart_state['restarts_completed']
        # If this is an integer, use the last N restarts
        elif self.restarts_to_use >= 1:
            last_N_restarts = self.restarts_to_use
        # If it's a decimal between 0 and 1, use it as a fraction
        # At restart 1, and a fraction of 0.5, this should just use restart 1
        elif 0 < self.restarts_to_use < 1:
            last_N_restarts = int(self.restarts_to_use * (1 + restart_state['restarts_completed']))

            # If this fraction is <1, use all until it's not
            if last_N_restarts < 1:
                last_N_restarts = 1 + restart_state['restarts_completed']

        log.debug(f"Last N is {last_N_restarts}")
        first_restart = max(1 + restart_state['restarts_completed'] - last_N_restarts, 0)
        usable_restarts = range(first_restart, 1 + restart_state['restarts_completed'])

        log.info(
            f"At restart {restart_state['restarts_completed']}, building haMSM using data from restarts {list(usable_restarts)}"
        )
        for restart_number in usable_restarts:
            for run_number in range(1, 1 + restart_state['runs_completed']):

                west_file_path = f"restart{restart_number}/run{run_number}/west.h5"
                marathon_west_files.append(west_file_path)

        log.debug(f"WESTPA datafile for analysis are {marathon_west_files}")

        #
        # If this is the first restart, check to see if you got any target state flux
        if restart_state['restarts_completed'] == 0:
            pass

            # Check to see if you got any target flux in ANY runs
            target_reached = False
            for west_file_path in marathon_west_files:
                if check_target_reached(west_file_path):
                    target_reached = True
                    break

            # If you reached the target, clean up from the extensions and then continue as normal
            # If  extension_iters is set to 0, then don't do extensions.
            if target_reached or self.extension_iters == 0:

                log.info("All runs reached target!")

                # Do some cleanup from the extension run
                if doing_extension and not self.extension_iters == 0:

                    # Remove the doing_extensions.lck lockfile
                    os.remove(EXTENSION_LOCKFILE)

                    westpa.rc.sim_manager.max_total_iterations = self.base_total_iterations

                # Otherwise, just continue as normal
                pass

            # If no runs reached the target, then we need to extend them
            elif not target_reached:

                log.info("Target not reached. Preparing for extensions.")

                # Create the doing_extensions.lck "lockfile" to indicate we're in extend mode (or keep if exists)
                #   and write the initial number of iterations to it.
                if not os.path.exists(EXTENSION_LOCKFILE):
                    with open(EXTENSION_LOCKFILE, 'w') as lockfile:
                        lockfile.write(str(self.max_total_iterations))

                # Reset runs_completed to 0, and rewrite restart.dat accordingly
                restart_state['runs_completed'] = 0

                self.prepare_extension_run(run_number=1, restart_state=restart_state, first_extension=True)
                return

        log.debug("Building haMSM and computing steady-state")
        log.debug(f"Cur iter is {self.cur_iter}")
        ss_dist, ss_flux, model = msmwe_compute_ss(self.plugin_config, marathon_west_files)
        self.ss_dist = ss_dist
        self.model = model

        log.debug(f'Steady-state distribution: {ss_dist}')
        log.info(f"Target steady-state flux is {ss_flux}")

        # Obtain cluster-structures
        log.debug("Obtaining cluster-structures")
        model.update_cluster_structures()

        # TODO: Do this with pathlib
        struct_directory = f"{restart_directory}/structs"
        if not os.path.exists(struct_directory):
            os.makedirs(struct_directory)

        flux_filename = f"{restart_directory}/JtargetSS.txt"
        with open(flux_filename, 'w') as fp:

            log.debug(f"Writing flux to {flux_filename}")
            fp.write(str(model.JtargetSS))
            fp.close()

        ss_filename = f"{restart_directory}/pSS.txt"
        with open(ss_filename, 'w') as fp:

            log.debug(f"Writing pSS to {ss_filename}")
            np.savetxt(fp, model.pSS)
            fp.close()

        # If this is the last run of the last restart, do nothing and exit.
        # if restart_state['runs_completed'] >= self.n_runs and restart_state['restarts_completed'] >= self.n_restarts:
        #     log.info("All restarts completed!")
        #     return

        # Construct start-state file with all structures and their weights
        # TODO: Don't explicitly write EVERY structure to disk, or this will be a nightmare for large runs.
        # However, for now, it's fine...
        log.debug("Writing structures")
        # TODO: Include start states from previous runs
        sstates_filename = f"{restart_directory}/startstates.txt"
        with open(sstates_filename, 'w') as fp:

            # Track the total number of segments iterated over
            seg_idx = 0

            log.info(f"Obtaining potential start structures ({len(model.cluster_structures.items())} bins avail)")

            # Can use these for sanity checks
            total_weight = 0.0
            total_bin_weights = []

            # Loop over each set of (bin index, all the structures in that bin)
            for (msm_bin_idx, structures) in tqdm.tqdm(model.cluster_structures.items()):

                total_bin_weights.append(0)

                # Don't put structures in the basis or target
                if msm_bin_idx in [model.n_clusters, model.n_clusters + 1]:
                    continue

                # The per-segment bin probability.
                # Map a cluster number onto a cluster INDEX, because after cleaning the cluster numbers may no longer
                # be consecutive.
                bin_prob = self.ss_dist[model.cluster_mapping[msm_bin_idx]]  # / len(structures)

                if bin_prob == 0:
                    log.info(f"MSM-Bin {msm_bin_idx}  has probability 0, so not saving any structs from it.")
                    continue

                # The total amount of WE weight in this MSM microbin
                msm_bin_we_weight = sum(model.cluster_structure_weights[msm_bin_idx])

                # Write each structure to disk. Loop over each structure within a bin.
                msm_bin_we_weight_tracker = 0
                for struct_idx, structure in enumerate(structures):

                    structure_filename = (
                        f"{struct_directory}/bin{msm_bin_idx}_" f"struct{struct_idx}.{STRUCT_EXTENSIONS[self.struct_filetype]}"
                    )

                    with self.struct_filetype(structure_filename, 'w') as struct_file:

                        # One structure per segment
                        seg_we_weight = model.cluster_structure_weights[msm_bin_idx][struct_idx]
                        msm_bin_we_weight_tracker += seg_we_weight

                        # Structure weights are set according to Algorithm 5.3 in
                        # Aristoff, D. & Zuckerman, D. M. Optimizing Weighted Ensemble Sampling of Steady States.
                        # Multiscale Model Sim 18, 646–673 (2020).
                        structure_weight = seg_we_weight * (bin_prob / msm_bin_we_weight)

                        total_bin_weights[-1] += structure_weight
                        total_weight += structure_weight

                        topology = model.reference_structure.topology

                        try:
                            angles = model.reference_structure.unitcell_angles[0]
                            lengths = model.reference_structure.unitcell_lengths[0] * 10
                        # This throws typeerror if reference_structure.unitcell_angles is None, or AttributeError
                        #   if reference_structure.unitcell_angles doesn't exist.
                        except (TypeError, AttributeError):
                            angles, lengths = None, None

                        coords = structure * 10  # Correct units

                        # Write the structure file
                        if self.struct_filetype is md.formats.PDBTrajectoryFile:
                            struct_file.write(coords, topology, modelIndex=1, unitcell_angles=angles, unitcell_lengths=lengths)

                        elif self.struct_filetype is md.formats.AmberRestartFile:
                            # AmberRestartFile takes slightly differently named keyword args
                            struct_file.write(coords, time=None, cell_angles=angles, cell_lengths=lengths)

                        else:
                            # Otherwise, YOLO just hope all the positional arguments are in the right place
                            log.warning(
                                f"This output filetype ({self.struct_filetype}) is probably supported, "
                                f"but not explicitly handled."
                                " You should ensure that it takes argument as (coords, topology)"
                            )
                            struct_file.write(coords, topology)
                            raise Exception("Don't know what extension to use for this filetype")

                        # Add this start-state to the start-states file
                        # This path is relative to WEST_SIM_ROOT
                        fp.write(f'b{msm_bin_idx}_s{struct_idx} {structure_weight} {structure_filename}\n')
                        seg_idx += 1

                # log.info(f"WE weight ({msm_bin_we_weight_tracker:.5e} / {msm_bin_we_weight:.5e})")

            # TODO: Fix this check. It's never quite worked right, nor has it ever caught an actual problem, so just
            #   disable for now.
            # In equilibrium, all probabilities count, but in steady-state the last 2 are the target/basis
            # Subtract off the probabilities of the basis and target states, since those don't have structures
            #   assigned to them.
            # assert np.isclose(total_weight, 1 - sum(model.pSS[model.n_clusters :])), (
            #     f"Total steady-state structure weights not normalized! (Total: {total_weight}) "
            #     f"\n\t pSS: {model.pSS}"
            #     f"\n\t Total bin weights {total_bin_weights}"
            #     f"\n\t pSS sum: {sum(model.pSS)}"
            #     f"\n\t pSS -2 sum: {sum(model.pSS[:-2])}"
            #     f"\n\t pSS (+target, no basis) sum: {sum(model.pSS[:-2]) + model.pSS[-1]}"
            # )

        ### Start the new simulation

        bstates_str = ""
        for original_bstate in original_bstates:
            orig_bstate_prob = original_bstate.probability
            orig_bstate_label = original_bstate.label
            orig_bstate_aux = original_bstate.auxref

            bstate_str = f"{orig_bstate_label} {orig_bstate_prob} {orig_bstate_aux}\n"

            bstates_str += bstate_str

        bstates_filename = f"{restart_directory}/basisstates.txt"
        with open(bstates_filename, 'w') as fp:
            fp.write(bstates_str)

        tstates_str = ""
        for original_tstate in original_tstates:
            orig_tstate_label = original_tstate.label
            # TODO: Handle multidimensional pcoords
            orig_tstate_pcoord = original_tstate.pcoord[0]

            tstate_str = f"{orig_tstate_label} {orig_tstate_pcoord}\n"
            tstates_str += tstate_str
        tstates_filename = f"{restart_directory}/targetstates.txt"
        with open(tstates_filename, 'w') as fp:
            fp.write(tstates_str)

        # Pickle the model
        objFile = f"{restart_directory}/hamsm.obj"
        with open(objFile, "wb") as objFileHandler:
            log.debug("Pickling model")
            pickle.dump(model, objFileHandler, protocol=4)
            objFileHandler.close()

        # Before finishing this restart, make a plot of the flux profile.
        #   This is made so the user can see whether

        self.generate_plots(restart_directory)

        # At this point, the restart is completed, and the data for the next one is ready (though still need to make the
        #   initialization file and such).

        if last_restart:
            log.info("All restarts completed! Finished.")
            return

        # Update restart_file file
        restart_state['restarts_completed'] += 1
        # If we're doing a restart, then reset the number of completed runs to 0 for the next marathon.
        restart_state['runs_completed'] = 0
        with open(self.restart_file, 'w') as fp:
            json.dump(restart_state, fp)

        log.info("Initializing new run")

        # TODO: Read this from config if available
        segs_per_state = 1

        old_initialization_path = self.initialization_file
        new_initialization_path = f"{restart_directory}/{self.initialization_file}"
        log.debug(f"Moving initialization file from {old_initialization_path} to {new_initialization_path}.")
        shutil.move(old_initialization_path, new_initialization_path)

        initialization_state = {
            'tstate_file': tstates_filename,
            'bstate_file': bstates_filename,
            'sstate_file': sstates_filename,
            'tstates': None,
            'bstates': None,
            'sstates': None,
            'segs_per_state': segs_per_state,
        }

        with open(self.initialization_file, 'w') as fp:
            json.dump(initialization_state, fp)

        westpa.rc.pstatus(
            f"\n\n"
            f"===== Restart {restart_state['restarts_completed']}, "
            + f"Run {restart_state['runs_completed']+1} initializing =====\n"
        )

        westpa.rc.pstatus(
            f"\nRun: \n\t w_init --tstate-file {tstates_filename} "
            + f"--bstate-file {bstates_filename} --sstate-file {sstates_filename} --segs-per-state {segs_per_state}\n"
        )

        w_init.initialize(**initialization_state, shotgun=False)

        log.info("New WE run ready!")
        westpa.rc.pstatus(f"\n\n===== Restart {restart_state['restarts_completed']} running =====\n")

        w_run.run_simulation()



'''Routines for configuring the work manager environment'''


import os
import re

from . import _available_work_managers


class WMEnvironment:
    '''A class to encapsulate the environment in which work managers are instantiated;
    this controls how environment variables and command-line arguments are used to
    set up work managers. This could be used to cleanly instantiate two work managers
    within one application, but is really more about providing facilities to make
    it easier for individual work managers to configure themselves according to
    precendence of configuration information:
      1. command-line arguments
      2. environment variables
      3. defaults
    '''

    group_title = 'parallelization options'
    group_description = None

    env_prefix = 'WM'
    arg_prefix = 'wm'

    default_work_manager = 'serial'
    default_parallel_work_manager = 'processes'
    valid_work_managers = list(_available_work_managers.keys())

    def __init__(self, use_arg_prefixes=False, valid_work_managers=None):
        self.environ = os.environ
        self.args = None

        # copy from the class variable to permit modification
        # this can be modified to disable certain work managers, for whatever reason
        # mostly it's about having a valid list for enumeration
        self.valid_work_managers = valid_work_managers or list(self.valid_work_managers)

        self.use_arg_prefixes = use_arg_prefixes

    def env_name(self, name):
        return '{}_{}'.format(self.env_prefix, name.upper())

    def arg_name(self, name):
        if self.use_arg_prefixes:
            return '{}_{}'.format(self.arg_prefix, name)
        else:
            return name

    def arg_flag(self, name):
        if self.use_arg_prefixes:
            return '--{}-{}'.format(self.arg_prefix, re.sub('_', '-', name))
        else:
            return '--{}'.format(re.sub('_', '-', name))

    def get_val(self, name, default=None, type_=None):
        envname = self.env_name(name)
        argname = self.arg_name(name)

        val = getattr(self.args, argname, None)
        if val is None:
            try:
                val = self.environ[envname]
            except KeyError:
                val = default

        if type_ is None:
            return val
        else:
            try:
                return type_(val)
            except ValueError as e:
                raise ValueError('cannot convert {!r} to {!r}: {!s}'.format(val, type_, e))

    def add_wm_args(self, parser):

        wm_group = parser.add_argument_group(self.group_title, self.group_description)
        wm_mutex = wm_group.add_mutually_exclusive_group()
        wm_mutex.add_argument(
            self.arg_flag('serial'),
            dest=self.arg_name('work_manager'),
            action='store_const',
            const='serial',
            help='run in serial mode',
        )
        wm_mutex.add_argument(
            self.arg_flag('parallel'),
            dest=self.arg_name('work_manager'),
            action='store_const',
            const=self.default_parallel_work_manager,
            help='run in parallel mode (using {})'.format(self.default_parallel_work_manager),
        )
        wm_mutex.add_argument(
            self.arg_flag('work_manager'),
            metavar='WORK_MANAGER',
            choices=self.valid_work_managers,
            help='''use the given work manager for parallel task distribution. Available
                                   work managers are {!r}; default is {!r}
                                   '''.format(
                tuple(self.valid_work_managers), self.default_work_manager
            ),
        )

        wm_group.add_argument(
            self.arg_flag('n_workers'),
            metavar='N_WORKERS',
            type=int,
            help='''Use up to N_WORKERS on this host, for work managers which support this option.
                                      Use 0 for a dedicated server. (Ignored by work managers which do not support
                                      this option.)''',
        )

        for wm in self.valid_work_managers:
            _available_work_managers[wm].add_wm_args(parser, self)

    def process_wm_args(self, args):
        self.args = args

    def make_work_manager(self):
        '''Using cues from the environment, instantiate a pre-configured work manager.'''

        work_manager_name = self.get_val('work_manager', '').lower()
        work_manager_name = work_manager_name or self.default_work_manager

        if work_manager_name not in self.valid_work_managers:
            raise ValueError('work manager {!r} is invalid or unavailable'.format(work_manager_name))
        else:
            return _available_work_managers[work_manager_name].from_environ(self)


default_env = WMEnvironment()
make_work_manager = default_env.make_work_manager
add_wm_args = default_env.add_wm_args
process_wm_args = default_env.process_wm_args



'''A system for parallel, remote execution of multiple arbitrary tasks.
Much of this, both in concept and execution, was inspired by (and in some
cases based heavily on) the ``concurrent.futures`` package from Python 3.2,
with some simplifications and adaptations (thanks to Brian Quinlan and his
futures implementation).
'''

import logging

from .core import WorkManager, WMFuture, FutureWatcher  # noqa


# Import core work managers, which should run most everywhere that
# Python does
from . import serial, threads, processes  # noqa
from .serial import SerialWorkManager
from .threads import ThreadsWorkManager
from .processes import ProcessWorkManager

log = logging.getLogger(__name__)

_available_work_managers = {'serial': SerialWorkManager, 'threads': ThreadsWorkManager, 'processes': ProcessWorkManager}

# Import ZeroMQ work manager if available
try:
    from . import zeromq  # noqa
    from .zeromq import ZMQWorkManager
except ImportError:
    log.info('ZeroMQ work manager not available')
    log.debug('traceback follows', exc_info=True)
else:
    _available_work_managers['zmq'] = ZMQWorkManager

# Import MPI work manager if available
try:
    from . import mpi  # noqa
    from .mpi import MPIWorkManager
except ImportError:
    log.info('MPI work manager not available')
    log.debug('traceback follows', exc_info=True)
else:
    _available_work_managers['mpi'] = MPIWorkManager

from . import environment  # noqa
from .environment import make_work_manager  # noqa


__all__ = [
    'serial',
    'threads',
    'processes',
    'SerialWorkManager',
    'ThreadsWorkManager',
    'ProcessWorkManager',
    'environment',
    'make_work_manager',
]
